#!/usr/bin/env python
"""
  Larch column file reader: read_ascii
"""
import os
import sys
import time
import string
from pathlib import Path
from collections import namedtuple
import numpy as np
from dateutil.parser import parse as dateparse
from math import log10
from larch import Group
from larch.symboltable import isgroup
from ..utils import read_textfile, format_exception, gformat, fix_varname
from .xafs_beamlines import guess_beamline

nanresult = namedtuple('NanResult', ('file_ok', 'message', 'nan_rows',
                                     'nan_cols', 'inf_rows', 'inf_cols'))

TINY = 1.e-7
MAX_FILESIZE = 100*1024*1024  # 100 Mb limit
COMMENTCHARS = '#;%*!$'

def look_for_nans(path):
    """
    look for Nans and Infs in an ASCII data file

    Arguments:
         path (string):  full path to ASCII column file

    Returns:
       NanResult, named tuple with elements

        'file_ok' : bool, whether data is read and contains no Nans or Infs
        'message' : exception message if file cannot be read at all or
                   'has nans', 'has infs' or 'has nans and infs'
        `nan_rows`: list of rows containing Nans
        `nan_cols`: list of columns containing Nans
        `inf_rows`: list of rows containing Infs
        `inf_cols`: list of columns containing Infs
    """

    nan_rows, nan_cols, inf_rows, inf_cols = [], [], [], []
    try:
        dat = read_ascii(path)
    except:
        print(''.join(format_exception()))
        return nanresult(False, f'could not read file {path}',
                             nan_rows, nan_cols, inf_rows, inf_cols)
    if len(dat.data) < 1:
        return nanresult(False, f'no data in file {path}',
                             nan_rows, nan_cols, inf_rows, inf_cols)
    if np.all(np.isfinite(dat.data)):
        return nanresult(True, 'file ok',
                             nan_rows, nan_cols, inf_rows, inf_cols)

    msg = 'unknown'
    nanvals = np.where(np.isnan(dat.data))
    if len(nanvals[0]) > 0:
        msg = 'has nans'
        for icol in nanvals[0]:
            if icol not in nan_cols:
                nan_cols.append(icol)
        for irow in nanvals[1]:
            if irow not in nan_rows:
                nan_rows.append(irow)

    infvals = np.where(np.isinf(dat.data))
    if len(infvals[0]) > 0:
        if len(msg) == 0:
            msg = 'has infs'
        else:
            msg = 'has nans and infs'
        for icol in infvals[0]:
            if icol not in inf_cols:
                inf_cols.append(icol)
        for irow in infvals[1]:
            if irow not in inf_rows:
                inf_rows.append(irow)

    return nanresult(False, msg, nan_rows, nan_cols, inf_rows, inf_cols)


def getfloats(txt, allow_times=True):
    """convert a line of numbers into a list of floats,
    as for reading a file with columnar numerical data.

    Arguments
    ---------
      txt   (str) : line of text to parse
      allow_times  (bool): whether to support time stamps [True]

    Returns
    -------
      list with each entry either a float or None

    Notes
    -----
      The `allow_times` will try to support common date-time strings
      using the dateutil module, returning a numerical value as the
      Unix timestamp, using
          time.mktime(dateutil.parser.parse(word).timetuple())
    """
    words = [w.strip() for w in txt.replace(',', ' ').split()]
    mktime = time.mktime
    for i, w in enumerate(words):
        val = None
        try:
            val = float(w)
        except ValueError:
            try:
                val = mktime(dateparse(w).timetuple())
            except ValueError:
                pass
        words[i] = val
    return words

def colname(txt):
    return fix_varname(txt.strip().lower()).replace('.', '_')


def parse_labelline(labelline, header):
    """
    parse the 'label line' for an ASCII file.


    This is meant to handle some special cases of XAFS data collected at a variety of sources
    """
    pass


def sum_fluor_channels(dgroup, roi, icr=None, ocr=None, ltime=None, label=None,
                       add_data=True, **kws):
    """build summed, deadtime-corrected fluorescence spectrum for a Group

    Arguments
    ---------
    dgroup    data group
    roi       list in array indices for ROI
    icr       None or list of array indices for ICR   [None]
    ocr       None or list of array indices for OCR   [None]
    ltime     None or list of array indices for LTIME [None]
    label     None or label for the summed, corrected array [None]
    add_data  bool, whether to add label and data to datgroup [False]

    Returns
    -------
    label, ndarray   with summed, deadtime-corrected data

    if add_data is True, the ndarray will also be appended to `dgroup.data,
    and the label will be appended to dgroup.array_labels


    Notes
    ------
    1.  The output array will be  Sum[ roi*icr/(ocr*ltime) ]
    2.  The default label will be like the array label for the 'dtcN' + first ROI
    3.  icr, ocr, or ltime can be `None`, '1.0', '-1', or '1' to mean '1.0' or
        arrays of indices for the respective components: must be the same lenght as roi

    4.  an array index of -1 will indicate 'bad channel' and be skipped for ROI
        or set to 1.0 for icr, ocr, or ltime

    5. if the list of arrays in roi, icr, ocr, or ltime are otherwise out-of-range,
       the returned (label, data) will be (None, None)

    """
    nchans = len(roi)
    if icr in ('1.0', -1, 1, None):
        icr = [-1]*nchans
    if ocr in ('1.0', -1, 1, None):
        ocr = [-1]*nchans
    if ltime in ('1.0', -1, 1, None):
        ltime = [-1]*nchans
    if len(ltime) != nchans or len(icr) != nchans or len(ocr) != nchans:
        raise Value("arrays of indices for for roi, icr, ocr, and ltime must be the same length")

    narr, npts = dgroup.data.shape
    nused = 0
    sum = 0.0
    olabel = None
    def get_data(arr, idx):
        iarr = arr[idx]
        if iarr < 0:
            return iarr, 1.0
        if iarr > narr-1:
            return None, None
        return iarr, dgroup.data[iarr, :]

    for pchan in range(nchans):
        droi = dicr = docr = dltime = 1.0
        iarr, droi = get_data(roi, pchan)
        if isinstance(droi, np.ndarray):
            if olabel is None:
                olabel = dgroup.array_labels[iarr]
        elif iarr is None:
            return (None, None)
        else:  # index of -1 here means "skip"
            continue

        iarr, dicr = get_data(icr, pchan)
        if iarr is None: return (None, None)

        iarr, docr = get_data(ocr, pchan)
        if iarr is None:  return (None, None)

        iarr, docr = get_data(ocr, pchan)
        if iarr is None:  return (None, None)

        iarr, dltime= get_data(ltime, pchan)
        if iarr is None:  return (None, None)

        sum += droi*dicr/(docr*dltime)
        nused += 1

    if label is None:
        if olabel is None: olabel = 'ROI'
        label = olabel = f'dtc{nused}_{olabel}'
        n  = 1
        while label in dgroup.array_labels:
            n += 1
            label = f'{olabel}_{n}'
    if add_data:
        dgroup.array_labels.append(label)
        dgroup.data = np.append(dgroup.data, sum.reshape(1, len(sum)), axis=0)
    return (label, sum)



def read_ascii(filename, labels=None, simple_labels=False,
               sort=False, sort_column=0):
    """read a column ascii column file, returning a group
    containing the data extracted from the file.

    Arguments:
      filename (str):        name of file to read
      labels (ist or None) : list of labels to use for array labels [None]
      simple_labels (bool) : whether to force simple column labels (note 1) [False]
      sort (bool) :          whether to sort row data (note 2) [False]
      sort_column (int) :    column to use for sorting (note 2) [0]

    Returns:
      Group

      A data group containing data read from file, with several attributes:

         | filename     : text name of the file.
         | array_labels : array labels, names of 1-D arrays.
         | data         : 2-dimensional data (ncolumns, nrows) with all data.
         | header       : array of text lines of the header.
         | footer       : array of text lines of the footer (text after the numerical data)
         | attrs        : group of attributes parsed from header lines.

    Notes:
      1. array labels.  If `labels` is `None` (the default), column labels
         and names of 1d arrays will be guessed from the file header.  This often
         means parsing the final header line, but tagged column files from several XAFS
         beamlines will be tried and used if matching.  Column labels may be like 'col1',
         'col2', etc if suitable column labels cannot be guessed.
         These labels will be used as names for the 1-d arrays from each column.
         If `simple_labels` is  `True`, the names 'col1', 'col2' etc will be used
         regardless of the column labels found in the file.

      2. sorting.  Data can be sorted to be in increasing order of any column,
         by giving the column index (starting from 0).

      3. header parsing. If header lines are of the forms of

           | KEY : VAL
           | KEY = VAL

         these will be parsed into a 'attrs' dictionary in the returned group.

    Examples:

        >>> feo_data = read_ascii('feo_rt1.dat')
        >>> show(g)a
        == Group ascii_file feo_rt1.dat: 0 methods, 8 attributes ==
        array_labels: ['energy', 'xmu', 'i0']
        attrs: <Group header attributes from feo_rt1.dat>
        data: array<shape=(3, 412), type=dtype('float64')>
        energy: array<shape=(412,), type=dtype('float64')>
        filename: 'feo_rt1.dat'
        header: ['# room temperature FeO', '# data from 20-BM, 2001, as part of NXS school', ... ]
        i0: array<shape=(412,), type=dtype('float64')>
        xmu: array<shape=(412,), type=dtype('float64')>

    See Also:
        read_xdi, write_ascii

    """
    if not Path(filename).is_file():
        raise OSError("File not found: '%s'" % filename)
    if os.stat(filename).st_size > MAX_FILESIZE:
        raise OSError("File '%s' too big for read_ascii()" % filename)

    text = read_textfile(filename)
    lines = text.split('\n')

    ncol = None
    data, footers, headers = [], [], []

    lines.reverse()
    section = 'FOOTER'

    for line in lines:
        line = line.strip()
        if len(line) < 1:
            continue
        # look for section transitions (going from bottom to top)
        if section == 'FOOTER' and not None in getfloats(line):
            section = 'DATA'
        elif section == 'DATA' and None in getfloats(line):
            section = 'HEADER'

        # act of current section:
        if section == 'FOOTER':
            footers.append(line)
        elif section == 'HEADER':
            headers.append(line)
        elif section == 'DATA':
            rowdat  = getfloats(line)
            if ncol is None:
                ncol = len(rowdat)
            elif ncol > len(rowdat):
                rowdat.extend([np.nan]*(ncol-len(rowdat)))
            elif ncol < len(rowdat):
                for i in data:
                    i.extend([np.nan]*(len(rowdat)-ncol))
                ncol = len(rowdat)
            data.append(rowdat)

    # reverse header, footer, data, convert to arrays
    footers.reverse()
    headers.reverse()
    data.reverse()
    data = np.array(data).transpose()

    # try to parse attributes from header text
    header_attrs = {}
    for hline in headers:
        hline = hline.strip().replace('\t', ' ')
        if len(hline) < 1: continue
        if hline[0] in COMMENTCHARS:
            hline = hline[1:].strip()
        keywds = []
        if ':' in hline: # keywords in  'x: 22'
            words = hline.split(':', 1)
            keywds = words[0].split()
        elif '=' in hline: # keywords in  'x = 22'
            words = hline.split('=', 1)
            keywds = words[0].split()
        if len(keywds) == 1:
            key = colname(keywds[0])
            if key.startswith('_'):
                key = key[1:]
            if len(words) > 1:
                header_attrs[key] = words[1].strip()


    fpath = Path(filename).absolute()
    filename = fpath.as_posix()
    attrs = {'filename': filename}
    group = Group(name='ascii_file %s' % filename,
                  path=filename, filename=fpath.name,
                  header=headers, data=[], array_labels=[])

    if len(data) == 0:
        return group

    if sort and sort_column >= 0 and sort_column < ncol:
         data = data[:, np.argsort(data[sort_column])]

    group.data = data

    if len(footers) > 0:
        group.footer = footers

    group.attrs = Group(name='header attributes from %s' % filename)
    for key, val in header_attrs.items():
        setattr(group.attrs, key, val)

    if isinstance(labels, str):
        for bchar in ',#@%|:*':
            labels = labels.replace(bchar, '')
        labels = labels.split()
    if labels is None and not simple_labels:
        bldat = guess_beamline(headers)(headers)
        labels = bldat.get_array_labels()

        if getattr(bldat, 'energy_units', 'eV') != 'eV':
            group.energy_units = bldat.energy_units
        if getattr(bldat, 'energy_column', 1) != 1:
            group.energy_column = bldat.energy_column
        if getattr(bldat, 'mono_dspace', -1) > 0:
            group.mono_dspace = bldat.mono_dspace

    set_array_labels(group, labels=labels, simple_labels=simple_labels)
    return group

def set_array_labels(group, labels=None, simple_labels=False,
                     save_oldarrays=False):

    """set array names for a group from its 2D `data` array.

    Arguments
    ----------
      labels (list of strings or None)  array of labels to use
      simple_labels (bool):   flag to use ('col1', 'col2', ...) [False]
      save_oldarrays (bool):  flag to save old array names [False]


    Returns
    -------
       group with newly named attributes of 1D array data, and
       an updated `array_labels` giving the mapping of `data`
       columns to attribute names.

    Notes
    ------
      1. if `simple_labels=True` it will overwrite any values in `labels`

      3. Array labels must be valid python names. If not enough labels
         are specified, or if name clashes arise, the array names may be
         modified, often by appending an underscore and letter or by using
         ('col1', 'col2', ...) etc.

      4. When `save_oldarrays` is `False` (the default), arrays named in the
         current `group.array_labels` will be erased.  Other arrays and
         attributes will not be changed.

    """
    write = sys.stdout.write
    if not hasattr(group, 'data'):
        write("cannot set array labels for group '%s': no `data`\n" % repr(group))
        return

    # clear old arrays, if desired
    oldlabels = getattr(group, 'array_labels', None)
    if oldlabels is not None and not save_oldarrays:
        for attr in oldlabels:
            if hasattr(group, attr):
                delattr(group, attr)

    ncols, nrow = group.data.shape

    ####
    # step 1: determine user-defined labels from input options
    # generating array `tlabels` for test labels
    #
    # generate simple column labels, used as backup
    clabels = ['col%d' % (i+1) for i in range(ncols)]

    if isinstance(labels, str):
        labels = labels.split()


    tlabels = labels
    # if simple column names requested or above failed, use simple column names
    if simple_labels or tlabels is None:
        tlabels = clabels

    ####
    # step 2: check input and correct problems
    # 2.a: check for not enough and too many labels
    if len(tlabels) < ncols:
        for i in range(len(tlabels), ncols):
            tlabels.append("col%i" % (i+1))
    elif len(tlabels) > ncols:
        tlabels = tlabels[:ncols]

    # 2.b: check for names that clash with group attributes
    # or that are repeated, append letter.
    reserved_names = ('data', 'array_labels', 'filename',
                      'attrs', 'header', 'footer')
    extras = string.ascii_lowercase
    labels = []
    for i in range(ncols):
        lname = tlabels[i]
        if lname in reserved_names or lname in labels:
            lname = lname + '_a'
            j = 0
            while lname in labels:
                j += 1
                if j == len(extras):
                    break
                lname = "%s_%s" % (tlabels[i], extras[j])
        if lname in labels:
            lname = clabels[i]
        labels.append(lname)

    ####
    # step 3: assign attribue names, set 'array_labels'
    for i, name in enumerate(labels):
        setattr(group, name, group.data[i])
    group.array_labels = labels
    return group


def write_ascii(filename, *args, commentchar='#', label=None, header=None):
    """
    write a list of items to an ASCII column file

    Arguments:
      args (list of groups):     list of groups to write.
      commentchar (str) :        character for comment ('#')
      label (str on None):       array label line (autogenerated)
      header (list of strings):  array of strings for header

    Returns:
      None

    Examples:
       >>> write_ascii('myfile',  group.energy, group.norm, header=['comment1', 'comment2']

    """
    ARRAY_MINLEN = 2
    write = sys.stdout.write
    com = commentchar
    label = label
    if header is None:
        header = []

    arrays = []
    arraylen = None

    for arg in args:
        if isinstance(arg, np.ndarray):
            if len(arg) > ARRAY_MINLEN:
                if arraylen is None:
                    arraylen = len(arg)
                else:
                    arraylen = min(arraylen, len(arg))
                arrays.append(arg)
            else:
                header.append(repr(arg))

        else:
            header.append(repr(arg))

    if arraylen is None:
        raise ValueError("write_ascii() need %i or more elements in arrays." % ARRAY_MINLEN)

    buff = []
    if header is None:
        buff = ['%s Output from Larch %s' % (com, time.ctime())]
    for s in header:
        buff.append('%s %s' % (com, s))
    buff.append('%s---------------------------------'% com)
    if label is None:
        label = (' '*13).join(['col%d' % (i+1) for i in range(len(arrays))])
    buff.append('#  %s' % label)

    arrays = np.array(arrays)
    for i in range(arraylen):
        w = [" %s" % gformat(val[i], length=14) for val in arrays]
        buff.append('  '.join(w))
    buff.append('')

    with open(filename, 'w', encoding=sys.getdefaultencoding()) as fout:
        fout.write('\n'.join(buff))
    sys.stdout.write("wrote to file '%s'\n" % filename)


def write_group(filename, group, scalars=None, arrays=None,
                arrays_like=None, commentchar='#'):
    """(deprecated) write components of a group to an ASCII column file


    Warning:
       This is pretty minimal and may work poorly for large groups of complex data.
       Use `save_session` instead.

    """

    items = dir(group)
    npts = 0
    if arrays is None:
        arrays = []
    if scalars is None:
        scalars = []

    if arrays_like is not None and arrays_like in items:
        array = getattr(group, arrays_like)
        if isinstance(array, np.ndarray):
            npts = len(array)

    for name in items:
        val = getattr(group, name)
        if isinstance(val, np.ndarray):
            if npts != 0 and npts == len(val) and name not in arrays:
                arrays.append(name)

    header =[]
    for s in scalars:
        if s in items:
            val = getattr(group, s)
            header.append("%s = %s" % (s, val))

    label = '  '.join(arrays)

    args = []
    for name in arrays:
        if name in items:
            args.append(getattr(group, name))

    write_ascii(filename, *args, commentchar=commentchar,
                label=label, header=header)


def read_fdmnes(filename, **kwargs):
    """read [FDMNES](http://fdmnes.neel.cnrs.fr/) ascii files"""
    group = read_ascii(filename, **kwargs)
    group.header_dict = dict(filetype='FDMNES', energy_units='eV')
    for headline in group.header:
        if ("E_edge" in headline):
            if headline.startswith("#"):
                headline = headline[1:]
            vals = [float(v) for v in headline.split(" = ")[0].split(" ") if v]
            vals_names = headline.split(" = ")[1].split(", ")
            group.header_dict.update(dict(zip(vals_names, vals)))
    group.name = f'FDMNES file {filename}'
    group.energy += group.header_dict["E_edge"]
    #fix _arrlabel -> arrlabel
    for ilab, lab in enumerate(group.array_labels):
        if lab.startswith("_"):
            fixlab = lab[1:]
            group.array_labels[ilab] = fixlab
            delattr(group, lab)
            setattr(group, fixlab, group.data[ilab])
    return group

def guess_filereader(path, return_text=False):
    """guess function name to use to read a data file based on the file header

    Arguments
    ---------
    path (str) : file path to be read

    Returns
    -------
    name of function (as a string) to use to read file
    if return_text: text of the read file
    """
    text = read_textfile(path)
    lines = text.split('\n')
    line1 = lines[0].lower()
    reader = 'read_ascii'
    if 'epics scan' in line1:
        reader = 'read_gsescan'
    if 'xdi' in line1:
        reader = 'read_xdi'
    if 'epics stepscan file' in line1 :
        reader = 'read_gsexdi'
    if ("#s" in line1) or ("#f" in line1):
        reader = 'read_specfile'
    if 'fdmnes' in line1:
        reader = 'read_fdmnes'
    if return_text:
        return reader, text
    else:
        return reader



"""Peak detection algorithm borrowed from PeakUtils

original author: Lucas Hermann Negri <lucashnegri@gmail.com>

"""
import numpy as np

def peak_indices(y, threshold=0.1, min_dist=1):
    """Peak detection routine.

    Finds the indices of peaks in `y` data by taking its first order
    derivative.  and using the `thres` and `min_dist` parameters, to
    control the number of peaks detected.

    Parameters
    ----------
    y : 1D ndarray
        1D amplitude data to search for peaks.
    threshold : float [0.1]
        Normalized peak threshold, as fraction of peak-to-peak of `y`.
        Peaks with amplitude higher than the threshold will be detected.
    min_dist: int [1]
        Minimum index distance between detected peaks. When multiple
        peaks are within this distance, the peak with highest amplitude
        is preferred.

    Returns
    -------
    ndarray
        Array with numeric indexes of the peaks that were detected.
    """
    thres = threshold * np.ptp(y) + y.min()
    min_dist = int(min_dist)

    # compute first order difference
    dy = np.diff(y)

    # propagate left and right values successively to fill all
    # plateau pixels (0-value)
    zeros, = np.where(dy == 0)

    # check if the signal is totally flat
    if len(zeros) == len(y) - 1:
        return np.array([])

    if len(zeros):
        # compute first order difference of zero indexes
        zeros_diff = np.diff(zeros)
        # check when zeros are not chained together
        zeros_diff_not_one, = np.add(np.where(zeros_diff != 1), 1)
        # make an array of the chained zero indexes
        zero_plateaus = np.split(zeros, zeros_diff_not_one)

        # fix if leftmost value in dy is zero
        if zero_plateaus[0][0] == 0:
            dy[zero_plateaus[0]] = dy[zero_plateaus[0][-1] + 1]
            zero_plateaus.pop(0)

        # fix if rightmost value of dy is zero
        if len(zero_plateaus) and zero_plateaus[-1][-1] == len(dy) - 1:
            dy[zero_plateaus[-1]] = dy[zero_plateaus[-1][0] - 1]
            zero_plateaus.pop(-1)

        # for each chain of zero indexes
        for plateau in zero_plateaus:
            median = np.median(plateau)
            # set leftmost values to leftmost non zero values
            dy[plateau[plateau < median]] = dy[plateau[0] - 1]
            # set rightmost and middle values to rightmost non zero values
            dy[plateau[plateau >= median]] = dy[plateau[-1] + 1]

    # find the peaks by using the first order difference
    peaks = np.where((np.hstack([dy, 0.0]) < 0.0) &
                     (np.hstack([0.0, dy]) > 0.0) &
                     (np.greater(y, thres)) )[0]

    # handle multiple peaks, respecting the minimum distance
    if peaks.size > 1 and min_dist > 1:
        highest = peaks[np.argsort(y[peaks])][::-1]
        rem = np.ones(y.size, dtype=bool)
        rem[peaks] = False

        for peak in highest:
            if not rem[peak]:
                sl = slice(max(0, peak - min_dist), peak + min_dist + 1)
                rem[sl] = True
                rem[peak] = False

        peaks = np.arange(y.size)[~rem]
    return peaks



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Plot RIXS data sets (2D maps)
================================
"""

import copy
import numpy as np
import matplotlib.pyplot as plt


from matplotlib import gridspec
from matplotlib import cm
from matplotlib.ticker import MaxNLocator, AutoLocator
from sympy import EX

from larch.utils.logging import getLogger

_logger = getLogger(__name__)  #: module logger, used as self._logger if not given


def plot_rixs(
    rd,
    et=True,
    fig_name="plot_rixs_fig",
    fig_size=(10, 10),
    fig_dpi=75,
    fig_title=None,
    x_label=None,
    y_label=None,
    x_nticks=0,
    y_nticks=0,
    x_min=None,
    x_max=None,
    y_min=None,
    y_max=None,
    cbar_show=True,
    cbar_pos="vertical",
    cbar_nticks=0,
    cbar_label="Signal intensity",
    cbar_norm0=False,
    cont_nlevels=50,
    cont_imshow=True,
    cmap=cm.gist_heat_r,
    cmap2=cm.RdBu,
    cmap_linlog="linear",
    cont_type="line",
    cont_lwidths=0.25,
    cont_labels=None,
    cont_labelformat="%.3f",
    origin="lower",
):
    """RIXS map plotter

    Parameters
    ----------

    rd : RixsData

    cbar_norm0 : boolean, optional [False]
        Normalize color bar around 0

    cont_levels : int, optional [50]
        number of contour lines

    cont_imshow : boolean, optional [True]
        use plt.imshow instead of plt.contourf

    """
    if not "RixsData" in str(type(rd)):
        _logger.error('only "RixsData" objects can be plotted!')
        return

    if fig_title is None:
        fig_title = rd.label

    if x_label is None:
        x_label = "Incoming energy (eV)"

    if et:
        try:
            x = rd.ene_in
            y = rd.ene_et
            zz = rd.rixs_et_map
            if y_label is None:
                y_label = "Energy transfer (eV)"
        except Exception:
            _logger.error("`ene_in/ene_et/rixs_et_map` arrays missing")
            return
    else:
        try:
            x = rd.ene_in
            y = rd.ene_out
            zz = rd.rixs_map
            if y_label is None:
                y_label = "Emitted energy (eV)"
        except Exception:
            _logger.error("`ene_in/ene_out/rixs_map` arrays missing")
            return

    plt.close(fig_name)
    fig = plt.figure(num=fig_name, figsize=fig_size, dpi=fig_dpi)

    # NOTE: np.nanmin/np.nanmax fails with masked arrays! better
    #       to work with MaskedArray for zz

    # if not 'MaskedArray' in str(type(zz)):
    #    zz = np.ma.masked_where(zz == np.nan, zz)

    # NOTE2: even with masked arrays min()/max() fail!!!  I do a
    #        manual check against 'nan' instead of the masked
    #        array solution

    try:
        zzmin, zzmax = np.nanmin(zz), np.nanmax(zz)
    except:
        zzmin, zzmax = np.min(zz), np.max(zz)

    if cbar_norm0:
        # normalize colors around 0
        if abs(zzmin) > abs(zzmax):
            vnorm = abs(zzmin)
        else:
            vnorm = abs(zzmax)
        norm = cm.colors.Normalize(vmin=-vnorm, vmax=vnorm)
    else:
        # normalize colors from min to max
        norm = cm.colors.Normalize(vmin=zzmin, vmax=zzmax)

    extent = (x.min(), x.max(), y.min(), y.max())
    levels = np.linspace(zzmin, zzmax, cont_nlevels)

    ### FIGURE LAYOUT ###
    plane = fig.add_subplot(111)
    plane.set_title(fig_title)
    plane.set_xlabel(x_label)
    plane.set_ylabel(y_label)
    if x_min and x_max:
        plane.set_xlim(x_min, x_max)
    if y_min and y_max:
        plane.set_ylim(y_min, y_max)

    # contour mode: 'contf' or 'imshow'
    if cont_imshow:
        contf = plane.imshow(zz, origin="lower", extent=extent, cmap=cmap, norm=norm)
    else:
        contf = plane.contourf(
            x, y, zz, levels, cmap=cm.get_cmap(cmap, len(levels) - 1), norm=norm
        )

    if "line" in cont_type.lower():
        cont = plane.contour(
            x, y, zz, levels, colors="k", hold="on", linewidths=cont_lwidths
        )
    if x_nticks:
        plane.xaxis.set_major_locator(MaxNLocator(int(x_nticks)))
    else:
        plane.xaxis.set_major_locator(AutoLocator())
    if y_nticks:
        plane.yaxis.set_major_locator(MaxNLocator(int(y_nticks)))
    else:
        plane.yaxis.set_major_locator(AutoLocator())

    # colorbar
    if cbar_show:
        xyratio = y.shape[0] / x.shape[0]
        cbar = fig.colorbar(
            contf,
            use_gridspec=True,
            orientation=cbar_pos,
            fraction=0.046 * xyratio,
            pad=0.04,
        )
        if cbar_nticks:
            cbar.set_ticks(MaxNLocator(int(y_nticks)))
        else:
            cbar.set_ticks(AutoLocator())
        cbar.set_label(cbar_label)

    fig.tight_layout()
    return fig


def plot_rixs_cuts(rd, et=True, fig_name="plot_rixs_cuts", fig_size=(8, 10), fig_dpi=75):
    """plot RIXS line cuts"""
    assert len(rd.line_cuts.keys()) >= 1, "no line cuts are present"
    plt.close(fig_name)
    fig, axs = plt.subplots(nrows=3, num=fig_name, figsize=fig_size, dpi=fig_dpi)

    for ax in axs:
        ax.set_axis_off()

    y_label = "Signal intensity"

    for key, val in rd.line_cuts.items():
        x, y, info = val["x"], val["y"], val["info"]
        mode = info["mode"]
        label = info["label"]
        color = info["color"]
        if mode == "CEE":
            ax = axs[0]
            ax.set_axis_on()
            x_label = "Incoming energy (eV)"
        elif mode == "CIE":
            ax = axs[1]
            ax.set_axis_on()
            if et:
                x = info['enecut'] - x
                x_label = "Energy transfer (eV)"
            else:
                x_label = "Emitted energy (eV)"
        elif mode == "CET":
            ax = axs[2]
            ax.set_axis_on()
            x_label = "Incoming energy (eV)"
        else:
            _logger.error(f"wrong mode: {mode}")
            continue
        ax.set_title(mode)
        ax.set_xlabel(x_label)
        ax.set_ylabel(y_label)
        ax.plot(x, y, label=label, color=color)
        ax.legend()
    fig.tight_layout()
    return fig


class RixsDataPlotter(object):
    """plotter for a RixsData object"""

    def __init__(self, rd):
        "initialize with keyword arguments dictionaries"
        if not "RixsData" in str(type(rd)):
            _logger.error('I can only plot "RixsData" objects!')
            return
        try:
            self.kwsd = copy.deepcopy(rd.kwsd["plot"])
        except Exception:
            self.kwsd = self.get_plot_kwsd()
        self.rd = rd

    def get_plot_kwsd(self):
        """return a dictionary of dictionaries with default keywords arguments"""
        kwsd = {
            "replace": True,
            "figname": "RixsDataPlotter",
            "figsize": (10, 10),
            "figdpi": 150,
            "title": None,
            "xlabel": None,
            "ylabel": None,
            "x_nticks": 0,
            "y_nticks": 0,
            "z_nticks": 0,
            "xlabelE": r"Incoming Energy (eV)",
            "ylabelE": r"Emitted Energy (eV)",
            "ylabelEt": r"Energy transfer (eV)",
            "zlabel": r"Intensity (a.u)",
            "xystep": 0.01,
            "xmin": None,
            "xmax": None,
            "ymin": None,
            "ymax": None,
            "xshift": 0,
            "ystack": 0,
            "xscale": 1,
            "yscale": 1,
            "cbar_show": False,
            "cbar_pos": "vertical",
            "cbar_nticks": 0,
            "cbar_label": "Counts/s",
            "cbar_norm0": False,
            "cmap": cm.gist_heat_r,
            "cmap2": cm.RdBu,
            "cmap_linlog": "linear",
            "cont_imshow": True,
            "cont_type": "line",
            "cont_lwidths": 0.25,
            "cont_levels": 50,
            "cont_labels": None,
            "cont_labelformat": "%.3f",
            "origin": "lower",
            "lcuts": False,
            "xcut": None,
            "ycut": None,
            "dcut": None,
            "lc_dticks": 2,
            "lc_color": "red",
            "lc_lw": 3,
        }
        return kwsd

    def plot(self, x=None, y=None, zz=None, **kws):
        """make the plot"""
        if x is None:
            x = self.rd.ene_in
            x0 = self.rd.ene_in
        if y is None:
            y = self.rd.ene_et
            y0 = self.rd.ene_out
        if zz is None:
            zz = self.rd.rixs_et_map
            zz0 = self.rd.rixs_map

        self.kwsd.update(**kws)

        # check if x and y are 1D or 2D arrays
        if (len(x.shape) == 1) and (len(y.shape) == 1):
            _xyshape = 1
        elif (len(x.shape) == 2) and (len(y.shape) == 2):
            _xyshape = 2

        lcuts = kws.get("lcuts", self.kwsd["lcuts"])
        xcut = kws.get("xcut", self.kwsd["xcut"])
        ycut = kws.get("ycut", self.kwsd["ycut"])
        dcut = kws.get("dcut", self.kwsd["dcut"])

        lc_dticks = kws.get("lc_dticks", self.kwsd["lc_dticks"])
        lc_color = kws.get("lc_color", self.kwsd["lc_color"])
        lc_lw = kws.get("lc_lw", self.kwsd["lc_lw"])

        replace = kws.get("replace", self.kwsd["replace"])
        figname = kws.get("figname", self.kwsd["figname"])
        figsize = kws.get("figsize", self.kwsd["figsize"])
        figdpi = kws.get("figdpi", self.kwsd["figdpi"])
        title = kws.get("title", self.kwsd["title"])
        xlabel = kws.get("xlabel", self.kwsd["xlabelE"])
        if y.max() / x.max() < 0.5:
            ylabel = kws.get("ylabel", self.kwsd["ylabelEt"])
        else:
            ylabel = kws.get("ylabel", self.kwsd["ylabelE"])
        zlabel = kws.get("zlabel", self.kwsd["zlabel"])
        xmin = kws.get("xmin", self.kwsd["xmin"])
        xmax = kws.get("xmax", self.kwsd["xmax"])
        ymin = kws.get("ymin", self.kwsd["ymin"])
        ymax = kws.get("ymax", self.kwsd["ymax"])
        x_nticks = kws.get("x_nticks", self.kwsd["x_nticks"])
        y_nticks = kws.get("y_nticks", self.kwsd["y_nticks"])
        z_nticks = kws.get("z_nticks", self.kwsd["z_nticks"])
        cmap = kws.get("cmap", self.kwsd["cmap"])

        cbar_show = kws.get("cbar_show", self.kwsd["cbar_show"])
        cbar_pos = kws.get("cbar_pos", self.kwsd["cbar_pos"])
        cbar_nticks = kws.get("cbar_nticks", self.kwsd["cbar_nticks"])
        cbar_label = kws.get("cbar_label", self.kwsd["cbar_label"])
        cbar_norm0 = kws.get("cbar_norm0", self.kwsd["cbar_norm0"])

        cont_imshow = kws.get("cont_imshow", self.kwsd["cont_imshow"])
        cont_type = kws.get("cont_type", self.kwsd["cont_type"])
        cont_levels = kws.get("cont_levels", self.kwsd["cont_levels"])
        cont_lwidths = kws.get("cont_lwidths", self.kwsd["cont_lwidths"])

        # NOTE: np.nanmin/np.nanmax fails with masked arrays! better
        #       to work with MaskedArray for zz

        # if not 'MaskedArray' in str(type(zz)):
        #    zz = np.ma.masked_where(zz == np.nan, zz)

        # NOTE2: even with masked arrays min()/max() fail!!!  I do a
        #        manual check against 'nan' instead of the masked
        #        array solution

        try:
            zzmin, zzmax = np.nanmin(zz), np.nanmax(zz)
        except:
            zzmin, zzmax = np.min(zz), np.max(zz)

        if cbar_norm0:
            # normalize colors around 0
            if abs(zzmin) > abs(zzmax):
                vnorm = abs(zzmin)
            else:
                vnorm = abs(zzmax)
            norm = cm.colors.Normalize(vmin=-vnorm, vmax=vnorm)
        else:
            # normalize colors from min to max
            norm = cm.colors.Normalize(vmin=zzmin, vmax=zzmax)

        extent = (x.min(), x.max(), y.min(), y.max())
        levels = np.linspace(zzmin, zzmax, cont_levels)

        ### FIGURE LAYOUT ###
        if replace:
            plt.close(figname)
        self.fig = plt.figure(num=figname, figsize=figsize, dpi=figdpi)
        if replace:
            self.fig.clear()

        # 1 DATA SET WITH OR WITHOUT LINE CUTS
        if lcuts:
            gs = gridspec.GridSpec(3, 3)  # 3x3 grid
            self.plane = plt.subplot(gs[:, :-1])  # plane
            self.lxcut = plt.subplot(gs[0, 2])  # cut along x-axis
            self.ldcut = plt.subplot(gs[1, 2])  # cut along d-axis (diagonal)
            self.lycut = plt.subplot(gs[2, 2])  # cut along y-axis
        else:
            self.plane = self.fig.add_subplot(111)  # plot olny plane

        # plane
        if title:
            self.plane.set_title(title)
        self.plane.set_xlabel(xlabel)
        self.plane.set_ylabel(ylabel)
        if xmin and xmax:
            self.plane.set_xlim(xmin, xmax)
        if ymin and ymax:
            self.plane.set_ylim(ymin, ymax)

        # contour mode: 'contf' or 'imshow'
        if cont_imshow:
            self.contf = self.plane.imshow(
                zz, origin="lower", extent=extent, cmap=cmap, norm=norm
            )
        else:
            self.contf = self.plane.contourf(
                x, y, zz, levels, cmap=cm.get_cmap(cmap, len(levels) - 1), norm=norm
            )

        if "line" in cont_type.lower():
            self.cont = self.plane.contour(
                x, y, zz, levels, colors="k", hold="on", linewidths=cont_lwidths
            )
        if x_nticks:
            self.plane.xaxis.set_major_locator(MaxNLocator(int(x_nticks)))
        else:
            self.plane.xaxis.set_major_locator(AutoLocator())
        if y_nticks:
            self.plane.yaxis.set_major_locator(MaxNLocator(int(y_nticks)))
        else:
            self.plane.yaxis.set_major_locator(AutoLocator())

        # colorbar
        if cbar_show:
            self.cbar = self.fig.colorbar(
                self.contf, use_gridspec=True, orientation=cbar_pos
            )
            if cbar_nticks:
                self.cbar.set_ticks(MaxNLocator(int(y_nticks)))
            else:
                self.cbar.set_ticks(AutoLocator())
            self.cbar.set_label(cbar_label)

        # xcut plot
        if lcuts and xcut:
            xpos = np.argmin(np.abs(xcut - x))
            if _xyshape == 1:
                self.lxcut.plot(
                    y, zz[:, xpos], label=str(x[xpos]), color=lc_color, linewidth=lc_lw
                )
            elif _xyshape == 2:
                self.lxcut.plot(
                    y[:, xpos],
                    zz[:, xpos],
                    label=str(x[:, xpos][0]),
                    color=lc_color,
                    linewidth=lc_lw,
                )
            if y_nticks:
                self.lxcut.xaxis.set_major_locator(
                    MaxNLocator(int(y_nticks / lc_dticks))
                )
            else:
                self.lxcut.xaxis.set_major_locator(AutoLocator())
            if z_nticks:
                self.lxcut.yaxis.set_major_locator(
                    MaxNLocator(int(z_nticks / lc_dticks))
                )
            else:
                self.lxcut.yaxis.set_major_locator(AutoLocator())
            self.lxcut.set_yticklabels([])
            self.lxcut.set_ylabel(zlabel)
            self.lxcut.set_xlabel(ylabel)
            if ymin and ymax:
                self.lxcut.set_xlim(ymin, ymax)

        # ycut plot
        if lcuts and ycut:
            ypos = np.argmin(np.abs(ycut - y))
            if _xyshape == 1:
                self.lycut.plot(
                    x, zz[ypos, :], label=str(y[ypos]), color=lc_color, linewidth=lc_lw
                )
            elif _xyshape == 2:
                self.lycut.plot(
                    x[ypos, :],
                    zz[ypos, :],
                    label=str(y[ypos, :][0]),
                    color=lc_color,
                    linewidth=lc_lw,
                )
            if x_nticks:
                self.lycut.xaxis.set_major_locator(
                    MaxNLocator(int(x_nticks / lc_dticks))
                )
            else:
                self.lycut.xaxis.set_major_locator(AutoLocator())
            if z_nticks:
                self.lycut.yaxis.set_major_locator(
                    MaxNLocator(int(z_nticks / lc_dticks))
                )
            else:
                self.lycut.yaxis.set_major_locator(AutoLocator())
            self.lycut.set_yticklabels([])
            self.lycut.set_ylabel(zlabel)
            self.lycut.set_xlabel(xlabel)
            if xmin and xmax:
                self.lycut.set_xlim(xmin, xmax)

        # dcut plot => equivalent to ycut plot for (zz0, x0, y0)
        if lcuts and dcut:
            ypos0 = np.argmin(np.abs(dcut - y0))
            if _xyshape == 1:
                self.ldcut.plot(
                    x0,
                    zz0[ypos0, :],
                    label=str(y0[ypos0]),
                    color=lc_color,
                    linewidth=lc_lw,
                )
            elif _xyshape == 2:
                self.ldcut.plot(
                    x0[ypos0, :],
                    zz0[ypos0, :],
                    label=str(y0[ypos0, :][0]),
                    color=lc_color,
                    linewidth=lc_lw,
                )
            if x_nticks:
                self.ldcut.xaxis.set_major_locator(
                    MaxNLocator(int(x_nticks / lc_dticks))
                )
            else:
                self.ldcut.xaxis.set_major_locator(AutoLocator())
            if z_nticks:
                self.ldcut.yaxis.set_major_locator(
                    MaxNLocator(int(z_nticks / lc_dticks))
                )
            else:
                self.ldcut.yaxis.set_major_locator(AutoLocator())
            self.ldcut.set_yticklabels([])
            self.ldcut.set_ylabel(zlabel)
            self.ldcut.set_xlabel(xlabel)
            if xmin and xmax:
                self.ldcut.set_xlim(xmin, xmax)
        plt.draw()
        plt.show()


if __name__ == "__main__":
    pass



#!/usr/bin/env python
"""

Basic unit testing utilities for larch

"""
import unittest
import time
import os
import sys
import numpy as np
from tempfile import NamedTemporaryFile
from larch import Interpreter, InputText

from io import StringIO

def nullfunction(*args, **kwargs):
    pass


class LarchSession(object):
    def __init__(self):
        self._larch = Interpreter()
        self.input  = self._larch.input
        #  InputText(prompt='test>', _larch=self._larch)
        self.symtable = self._larch.symtable
        setsym = self.symtable.set_symbol
        setsym('testdir',  os.getcwd())
        setsym('_plotter.no_plotting', True)
        setsym('_plotter.get_display',  nullfunction)
        self.set_stdout()

    def set_stdout(self, fname='_stdout_'):
        # self._outfile = os.path.abspath(fname)
        # self._larch.writer = open(self._outfile, 'w')
        self._larch.writer = StringIO()

    def read_stdout(self):
        self._larch.writer.flush()
        self._larch.writer.seek(0)
        t0 = time.time()
        time.sleep(0.01)

        out = self._larch.writer.read()
        self._larch.writer.close()
        self.set_stdout()
        return out

    def run(self, text):
        return self._larch.eval(text, fname='test', lineno=0)

    def get_errors(self):
        return self._larch.error

    def get_symbol(self, name):
        return self.symtable.get_symbol(name, create=False)

class TestCase(unittest.TestCase):
    '''testing of larch'''
    def setUp(self):
        self.session = LarchSession()
        self.symtable = self.session.symtable

    def runscript(self, fname, dirname='.'):
        origdir = os.path.abspath(os.getcwd())
        dirname = os.path.abspath(dirname)
        os.chdir(dirname)
        with open(fname, 'r') as fh:
            text = fh.read()
        self.session.run(text)
        os.chdir(origdir)

    def trytext(self, text):
        self.session.set_stdout()
        ret = self.session.run(text)
        out = self.session.read_stdout()
        err = self.session.get_errors()
        return out, err

    def tearDown(self):
        pass

    def getSym(self, sym):
        return self.session.get_symbol(sym)

    def isValue(self, sym, val):
        '''assert that a symboltable symbol has a particular value'''
        testval = self.session.get_symbol(sym)
        if isinstance(val, np.ndarray):
            return self.assertTrue(np.all(testval == val))
        else:
            return self.assertTrue(testval == val)

    def isNear(self, expr, val, places=7):
        '''assert that a symboltable symbol is near a particular value'''
        testval = self.session.run(expr)
        if isinstance(val, np.ndarray):
            for x, y in zip(testval, val):
                self.assertAlmostEqual(x, y, places=places)
        else:
            return self.assertAlmostEqual(testval, val, places=places)

    def isTrue(self, expr):
        '''assert that an expression evaluates to True'''
        testval = self.session.run(expr)
        if isinstance(testval, np.ndarray):
            testval = np.all(testval)
        return self.assertTrue(testval)

    def isFalse(self, expr):
        '''assert that an expression evaluates to False'''
        testval = self.session.run(expr)
        if isinstance(testval, np.ndarray):
            testval = np.all(testval)
        return self.assertFalse(testval)

    def ExceptionRaised(self):
        return self.assertTrue(len(self.session.get_errors()) > 0)

    def NoExceptionRaised(self):
        return self.assertTrue(len(self.session.get_errors()) == 0)



#!/usr/bin/env python

## Autobk (XAFS background subtraction) in pure Python,
## using Python code from Lxsarch.
from larch import Interpreter

from larch_plugins.xafs import pre_edge, autobk
from larch_plugins.io import read_ascii

# create plain interpreter, don't load all the plugins
_larch = Interpreter(with_plugins=False)

fname = '../xafsdata/cu_rt01.xmu'

cu = read_ascii(fname, labels='energy mu i0', _larch=_larch)
print( 'Read ASCII File:', cu)
print( dir(cu))

pre_edge(cu, _larch=_larch)
print( 'After pre-edge:')
print( dir(cu))

autobk(cu, rbkg=1.0, kweight=1, _larch=_larch)

print( 'After autobk:')
print( dir(cu))



#!/usr/bin/env python
""" Larch Tests Version 1 """
import unittest
import time
import ast
import numpy as np
from sys import version_info

from utils import TestCase
from larch import Interpreter

class TestEval(TestCase):
    '''testing of ast evaluation'''

    def test_function1(self):
        "test function definition and running"
        self.trytext("""
def fcn(x, scale=2):
    'test function'
    out = sqrt(x)
    if scale > 1:
        out = out * scale
    endif
    return out
enddef
""")
        self.trytext("a = fcn(4, scale=9)")
        self.isValue("a", 18)
        self.NoExceptionRaised()

        self.trytext("a = fcn(9, scale=0)")
        self.isValue("a", 3)
        self.NoExceptionRaised()

        out, err = self.trytext("print(fcn)")
        self.NoExceptionRaised()
        out = out.split('\n')
        self.assertTrue(out[0].startswith('<Procedure fcn'))

        out, err = self.trytext("a = fcn()")
        self.ExceptionRaised()
        errtype, errmsg = err[0].get_error()

        errlines = errmsg.split('\n')

        self.assertTrue(errtype == 'TypeError')

        out, err = self.trytext("a = fcn(x, bogus=3)")
        self.ExceptionRaised()
        errtype, errmsg = err[0].get_error()
        errmsgs = errmsg.split('\n')
        self.assertTrue(errtype == 'NameError')

    def test_function_vararg(self):
        "test function with var args"
        self.trytext("""
def fcn(*args):
    'test varargs function'
    out = 0
    for i in args:
        out = out + i*i
    endfor
    return out
enddef
""")
        self.NoExceptionRaised()
        self.trytext("o = fcn(1,2,3)")
        self.NoExceptionRaised()
        self.isValue('o', 14)
        out, err = self.trytext("print(fcn)")
        out = out.split('\n')
        self.assertTrue(out[0].startswith('<Procedure fcn'))

    def test_function_kwargs(self):
        "test function with kw args, no **kws"
        self.trytext("""
def fcn(square=False, x=0, y=0, z=0, t=0):
    'test varargs function'
    out = 0
    for i in (x, y, z, t):
        if square:
            out = out + i*i
        else:
            out = out + i
        endif
    endfor
    return out
enddef
""")
        self.NoExceptionRaised()
        out, err = self.trytext("print(fcn)")
        self.NoExceptionRaised()
        out = out.split('\n')
        self.assertTrue(out[0].startswith('<Procedure fcn'))

        self.trytext("o = fcn(x=1, y=2, z=3, square=False)")
        self.NoExceptionRaised()
        self.isValue('o', 6)

        self.trytext("o = fcn(x=1, y=2, z=3, square=True)")
        self.NoExceptionRaised()
        self.isValue('o', 14)

        self.trytext("o = fcn(x=1, y=2, z=3, t=-2)")
        self.NoExceptionRaised()
        self.isValue('o', 4)

        out, err = self.trytext("o = fcn(x=1, y=2, z=3, t=-12, s=1)")
        self.ExceptionRaised()
        errtype, errmsg = err[0].get_error()
        self.assertTrue(errtype == 'TypeError')
        self.assertTrue('extra keyword arg' in errmsg)

    def test_function_kwargs1(self):
        "test function with **kws arg"
        self.trytext("""
def fcn(square=False, **kws):
    'test varargs function'
    out = 0
    for i in kws.values():
        if square:
            out = out + i*i
        else:
            out = out + i
        endif
    endfor
    return out
enddef
""")

        out, err = self.trytext("print(fcn)")
        out = out.split('\n')
        self.assertTrue(out[0].startswith('<Procedure fcn'))
        self.trytext("o = fcn(x=1, y=2, z=3, square=False)")
        self.isValue('o', 6)
        self.NoExceptionRaised()

        self.trytext("o = fcn(x=1, y=2, z=3, square=True)")
        self.isValue('o', 14)
        self.NoExceptionRaised()

    def test_function_kwargs2(self):
        "test function with positional and **kws args"
        self.trytext("""
def fcn(x, y, **kws):  # end of line comment
    'test function'
    if 'scale' in kws:  # optional scale
        scale = kws['scale']
    else:
        scale = 1
    #endif
    return scale*(x + y**2)
#enddef
""")
        out, err = self.trytext("print(fcn)")
        self.NoExceptionRaised()
        out = out.split('\n')
        self.assertTrue(out[0].startswith('<Procedure fcn'))

        self.trytext("o = -1")
        self.NoExceptionRaised()
        self.trytext("o = fcn(2, 1)")
        self.isValue('o', 3)
        self.NoExceptionRaised()

        self.trytext("o = fcn(x=1, y=2)")
        self.isValue('o', 5)
        self.NoExceptionRaised()

        self.trytext("o = fcn(y=2, x=7)")
        self.isValue('o', 11)
        self.NoExceptionRaised()

        self.trytext("o = fcn(1, y=2)")
        self.isValue('o', 5)
        self.NoExceptionRaised()

        self.trytext("o = fcn(1, y=2, scale=0)")
        self.isValue('o', 0)
        self.NoExceptionRaised()

        self.trytext("o = fcn(1, y=2, scale=2)")
        self.isValue('o', 10)
        self.NoExceptionRaised()

        self.trytext("o = fcn(1, y=2, unused_opt=1)")
        self.isValue('o', 5)
        self.NoExceptionRaised()

        out, err = self.trytext("o = fcn(1, x=2)")
        self.ExceptionRaised()
        errtype, errmsg = err[0].get_error()
        self.assertTrue(errtype == 'TypeError')


if __name__ == '__main__':  # pragma: no cover
    for suite in (TestEval,):
        suite = unittest.TestLoader().loadTestsFromTestCase(suite)
        unittest.TextTestRunner(verbosity=2).run(suite)



import glob
from gzip import GzipFile

from larch.io import read_athena

def get_athena_version(fname):
    ftype = 'ascii'
    try:
        fh = GzipFile(fname)
        text = fh.read()
        ftype = 'gzip'
    except:
        text = None

    if text is None:
        text = open(fname, 'r').read()

    if isinstance(text, bytes):
        text = text.decode('utf-8')
    version = 'unknown'
    for line in text[:200].split('\n'):
        line = line.lower().replace('#', '').replace('--', '').strip()
        if 'athena project file' in line:
            line = line.replace('version', '')
            line = line.replace('_____header1', '')
            line = line.replace('athena project file', '')
            words = [a.strip() for a in line.split()]
            version = words.pop().replace('"', "").replace(',', "")
            parent = words.pop()
    return (ftype, parent, version)


OutFormat = "{:30s} {:8s} {:12s} {:s}"
print(OutFormat.format('FileName', 'Type', 'Version', '# Groups'))
print('-'*60)

for fname in glob.glob('*.prj'):
    ftype, parent, version = get_athena_version(fname)
    prj = read_athena(fname, do_preedge=True, do_bkg=True, do_fft=True)
    n = "{:3d}".format(len(prj._athena_groups.keys()))
    print(OutFormat.format(fname, ftype, version, n))



#!/usr/bin/env python
""" Tests of Larch Scripts  """

from pathlib import Path
from utils import TestCase

class TestScripts(TestCase):
    '''test read_ascii() for all example xafsdata'''
    def test_read_ascii(self):
        # self.symtable.set_symbol('testdir',  os.getcwd())
        dirname = Path(__file__).parent.resolve() / 'larch_scripts'
        self.runscript('read_ascii.lar', dirname=dirname)
        assert(len(self.session.get_errors()) == 0)

        actual = self.session.get_symbol('results')
        expected = self.session.get_symbol('expected')
        print("actual ", actual)
        print("expected ", expected)
        
        for fname, ncol, nrow, labels in expected:
            acol, arow, alabs = actual[fname]
            assert(acol == ncol)
            assert(arow == nrow)
            assert(alabs == labels)




"""
main Larch Applications
"""
import sys
import locale
import inspect
import shutil
from argparse import ArgumentParser
from  pathlib import Path

import matplotlib
from pyshortcuts import make_shortcut, ico_ext, get_desktop

from .site_config import icondir, uname, update_larch
from .version import __date__, make_banner, check_larchversion

HAS_WXPYTHON = False
try:
    import wx
    HAS_WXPYTHON = True
except ImportError:
    pass

if HAS_WXPYTHON:
    # note: this will be needed for some macOS builds until wxPython 4.2.1 is released.
    if uname == 'darwin':
        wx.PyApp.IsDisplayAvailable = lambda _: True

def use_mpl_wxagg():
    """import matplotlib, set backend to wxAgg"""
    if HAS_WXPYTHON:
        try:
            matplotlib.use('WXAgg', force=True)
        except ImportError:
            pass

def set_locale():
    """set locale to 'C' for these applications"""
    locale.setlocale(locale.LC_ALL, 'C')

class LarchApp(object):
    """wrapper for Larh application"""
    def __init__(self, name, script, icon=None, description=None,
                 is_wxapp=True, filetype=None):
        self.name = name
        self.script = script
        self.is_wxapp = is_wxapp
        self.description = description or name
        self.icon = icon or 'larch'
        self.filetype = filetype or 'data file'

    def make_desktop_shortcut(self, folder='Larch'):
        """make (or remake) desktop shortcuts for Larch apps"""
        bindir = 'Scripts' if uname == 'win' else 'bin'
        bindir = Path(sys.prefix, bindir).absolute()
        script = self.script
        if not self.script.startswith('_'):
            script = Path(bindir, self.script).absolute().as_posix()

        icon = Path(icondir, self.icon).absolute()
        if isinstance(ico_ext, (list, tuple)):
            for ext in ico_ext:
                ticon = Path(f"{self.icon:s}.{ext:s}").absolute()
                if ticon.exists():
                    icon = ticon
        make_shortcut(script, name=self.name, folder=folder,
                      icon=icon.as_posix(),
                      description=self.description,
                      terminal=(not self.is_wxapp))


    def prep_cli(self):
        parser = ArgumentParser(description=self.description)
        parser.add_argument('filename', nargs='?',  help=self.filetype)

        parser.add_argument('-m', '-mode', dest='run_mode', action='store_true',
                            default='xas', help='set startup mode')
        parser.add_argument('-w', '-wx_inspect', dest='wx_inspect', action='store_true',
                            default=False, help='enable wxPython inspection and debugging')

        args = parser.parse_args()
        self.filename = None
        if 'filename' in args and args.filename is not None:
            self.filename = Path(args.filename).absolute()
        self.wx_inspect = args.wx_inspect
        self.run_mode = args.run_mode
        if self.is_wxapp:
            set_locale()
            use_mpl_wxagg()


# #             App Name,       icon,        terminal,  Script / pyshortcuts command, Description
# MainApps = (('Larch CLI',     'larch',       True,  'larch', 'Basic Command-line interface for Larch'),
#             ('Larch Updater', 'larch',       True,  '_ -m pip install --upgrade xraylarch', 'Larch Updatar'),
#             ('Larch GUI',     'larch',       False, 'larch --wxgui', 'Enhanced Command-line interface for Larch'),
#             ('XAS Viewer',    'onecone',     False, 'larix', 'XANES and EXAFS Analysis GUI for Larch'),
#             ('Larix',         'onecone',     False, 'larix', 'XANES and EXAFS Analysis GUI for Larch'),
#             ('GSE MapViewer', 'gse_xrfmap',  False, 'gse_mapviewer', 'XRF Map Viewing and Analysis'),
#             ('XRF Viewer',    'ptable',      False, 'larch_xrf', 'X-ray FluorescenceData Viewing and Analysis'),
#             ('XRD1D Viewer',  'larch',       False, 'larch_xrd1d', 'X-ray Diffraction Data Viewing'),
#             )
#


LarchApps = {
    'larch': LarchApp(name='Larch CLI', script='larch', icon='larch',
                      description='Basic Command-line interface for Larch'),
    'Larch GUI': LarchApp(name='Larch GUI', script='larch --wxgui', icon='larch',
                      description='Enhanced Command-line interface for Larch'),
    'Larch Updater': LarchApp(name='Update Larch',
                              script='_ -m pip install --upgrade xraylarch',
                              icon='larch',
                              description='Larch Updater', is_wxapp=False),

    'Larix': LarchApp(name='Larix', script='larix', icon='onecone',
                      description='XANES and EXAFS Analysis GUI for Larch'),
    'XRFMap Viewer': LarchApp(name='XRFMap Viewer', script='gse_mapviewer',
                              icon='gse_xrfmap', filetype='XRM Map File (.h5)',
                              description='XRFMap Viewing and Analysis'),
    'XRF Viewer': LarchApp(name='XRF Viewer', script='larch_xrf', icon='ptable',
                           description='X-ray FluorescenceData Viewing and Analysis'),
    'XRD1D Viewer': LarchApp(name='XRD1D Viewer', script='larch_xrd1d', icon='larch',
                             description='X-ray Diffraction Data Viewing'),
    }


# entry points:
def run_gse_mapviewer():
    "XRFMap Viewer"
    app = LarchApps['XRFMap Viewer']
    app.prep_cli()
    from .wxmap import MapViewer
    MapViewer(check_version=True, title=app.description,
              filename=app.filename).MainLoop()


def run_larix():
    """XANES and EXAFS Analysis GUI for Larch"""
    app = LarchApps['Larix']
    app.prep_cli()
    from .wxxas import LarixApp
    LarixApp(check_version=True, filename=app.filename,
             mode=app.run_mode, with_wx_inspect=app.wx_inspect).MainLoop()

run_xas_viewer = run_larix

def run_larch_xrf():
    """X-ray FluorescenceData Viewing and Analysis"""
    app = LarchApps['XRF Viewer']
    app.prep_cli()
    from .wxlib.xrfdisplay import XRFApp
    XRFApp(filename=app.filename).MainLoop()

def run_epics_xrf():
    """XRF Viewing and Control for Epics XRF Detectors"""
    app = LarchApps['XRF Viewer']
    app.prep_cli()
    try:
        from .epics import EpicsXRFApp
        EpicsXRFApp().MainLoop()
    except ImportError:
        print('cannot import EpicsXRFApp: try `pip install "xraylarch[epics]"`')

def run_larch_xrd1d():
    """X-ray Diffraction Data Display"""
    app = LarchApps['XRD1D Viewer']
    app.prep_cli()
    from .wxxrd import XRD1DApp
    XRD1DApp().MainLoop()

def run_xrd2d_viewer():
    """XRD Display for 2D patternss"""
    set_locale()
    use_mpl_wxagg()
    from .wxxrd import XRD2DViewer
    XRD2DViewer().MainLoop()

def run_gse_dtcorrect():
    """GSE DT Correct """
    set_locale()
    use_mpl_wxagg()
    from .wxmap import DTViewer
    DTViewer().MainLoop()


def run_feff6l():
    "run feff6l"
    from .xafs.feffrunner import feff6l_cli
    feff6l_cli()

def run_feff8l():
    "run feff8l"
    from .xafs.feffrunner import feff8l_cli
    feff8l_cli()

def run_larch_server():
    "run larch XMLRPC server"
    from .xmlrpc_server import larch_server_cli
    larch_server_cli()


## main larch cli or wxgui
def run_larch():
    """
    main larch application launcher, running either
    commandline repl program or wxgui
    """
    parser = ArgumentParser(description='run main larch program')
    sargs = (("-v", "--version", "version", False, "show version"),
             ("-e", "--exec", "noshell", False, "execute script only"),
             ("-q", "--quiet", "quiet", False, "set quiet mode"),
             ("-x", "--nowx", "nowx", False, "set no wx graphics mode"),
             ("-w", "--wxgui", "wxgui", False, "run Larch GUI"),
             ("-m", "--makeicons", "makeicons", False, "create desktop icons"),
             ("-u", "--update", "update", False, "update larch to the latest version"),
             ("-r", "--remote", "server_mode", False, "run in remote server mode"),
             ("-p", "--port", "port", "4966", "port number for remote server"))

    for opt, longopt, dest, default, help in sargs:
        parser.add_argument(opt, longopt, dest=dest, action='store_true',
                            default=default, help=help)

    parser.add_argument('scripts', nargs='*',
                        help='larch or python scripts to run on startup')

    args = parser.parse_args()
    if args.version:
        print(make_banner(with_libraries=True))
        vinfo = check_larchversion()
        if vinfo.update_available:
            print(vinfo.message)
        return

    with_wx = HAS_WXPYTHON and (not args.nowx)

    # create desktop icons
    if args.makeicons:
        larchdir = Path(get_desktop(), 'Larch').absolute()
        if Path(larchdir).exists():
            shutil.rmtree(larchdir)

        for n, app in LarchApps.items():
            app.make_desktop_shortcut()
        return

    # run updates
    if args.update:
        update_larch()
        return

    # run in server mode
    if args.server_mode:
        if with_wx:
            use_mpl_wxagg()
        vinfo = check_larchversion()
        if vinfo.update_available:
            print(vinfo.message)
        from .xmlrpc_server import LarchServer
        server = LarchServer(host='localhost', port=int(args.port))
        server.run()

    # run wx Larch GUI
    elif args.wxgui:
        set_locale()
        use_mpl_wxagg()
        from .wxlib.larchframe import LarchApp
        LarchApp(with_inspection=True).MainLoop()

    # run wx Larch CLI
    else:
        if with_wx:
            set_locale()
            use_mpl_wxagg()
        vinfo = check_larchversion()
        if vinfo.update_available:
            print(vinfo.message)

        from .shell import Shell
        cli = Shell(quiet=args.quiet, with_wx=with_wx)
        # execute scripts listed on command-line
        if args.scripts is not None:
            for script in args.scripts:
                if script.endswith('.py'):
                    cmd = f"import {script[:-3]}"
                else:
                    cmd = f"run('{script}')"
                cli.default(cmd)
        # if interactive, start command loop
        if not args.noshell:
            try:
                cli.cmdloop()
            except ValueError:
                pass



# -*- coding: utf-8 -*-
import os

from docutils.io import FileOutput
from docutils.frontend import OptionParser
from docutils import nodes

from docutils.utils import smartquotes

import sphinx.builders.latex
# from sphinx.util.smartypants import educate_quotes_latex
from sphinx.writers.latex import LaTeXWriter
from sphinx.util.console import bold
from sphinx.util.osutil import copyfile
from sphinx.util.texescape import tex_escape_map
import sphinx.writers.latex

# remove usepackage for sphinx here, we add it later in the preamble in conf.py
# sphinx.writers.latex.HEADER = sphinx.writers.latex.HEADER.replace('\usepackage{sphinx}', '')

import re


# Constants for quote education.
punct_class = r"""[!"#\$\%'()*+,-.\/:;<=>?\@\[\\\]\^_`{|}~]"""
end_of_word_class = r"""[\s.,;:!?)]"""
close_class = r"""[^\ \t\r\n\[\{\(\-]"""
dec_dashes = r"""&#8211;|&#8212;"""

# Special case if the very first character is a quote
# followed by punctuation at a non-word-break. Close the quotes by brute force:
single_quote_start_re = re.compile(r"""^'(?=%s\\B)""" % (punct_class,))
double_quote_start_re = re.compile(r"""^"(?=%s\\B)""" % (punct_class,))

# Special case for double sets of quotes, e.g.:
#   <p>He said, "'Quoted' words in a larger quote."</p>
double_quote_sets_re = re.compile(r""""'(?=\w)""")
single_quote_sets_re = re.compile(r"""'"(?=\w)""")

# Special case for decade abbreviations (the '80s):
decade_abbr_re = re.compile(r"""\b'(?=\d{2}s)""")

# Get most opening double quotes:
opening_double_quotes_regex = re.compile(r"""
                (
                        \s          |   # a whitespace char, or
                        &#160;      |   # a non-breaking space entity, or
                        --          |   # dashes, or
                        &[mn]dash;  |   # named dash entities
                        %s          |   # or decimal entities
                        &\#x201[34];    # or hex
                )
                "                 # the quote
                (?=\w)            # followed by a word character
                """ % (dec_dashes,), re.VERBOSE)

# Double closing quotes:
closing_double_quotes_regex = re.compile(r"""
                #(%s)?   # character that indicates the quote should be closing
                "
                (?=%s)
                """ % (close_class, end_of_word_class), re.VERBOSE)

closing_double_quotes_regex_2 = re.compile(r"""
                (%s)   # character that indicates the quote should be closing
                "
                """ % (close_class,), re.VERBOSE)

# Get most opening single quotes:
opening_single_quotes_regex = re.compile(r"""
                (
                        \s          |   # a whitespace char, or
                        &#160;      |   # a non-breaking space entity, or
                        --          |   # dashes, or
                        &[mn]dash;  |   # named dash entities
                        %s          |   # or decimal entities
                        &\#x201[34];    # or hex
                )
                '                 # the quote
                (?=\w)            # followed by a word character
                """ % (dec_dashes,), re.VERBOSE)

closing_single_quotes_regex = re.compile(r"""
                (%s)
                '
                (?!\s | s\b | \d)
                """ % (close_class,), re.VERBOSE)

closing_single_quotes_regex_2 = re.compile(r"""
                (%s)
                '
                (\s | s\b)
                """ % (close_class,), re.VERBOSE)


def educate_quotes_latex(s, dquotes=("``", "''")):
    """
    Parameter:  String.

    Returns:    The string, with double quotes corrected to LaTeX quotes.

    Example input:  "Isn't this fun?"
    Example output: ``Isn't this fun?'';
    """

    # Special case if the very first character is a quote
    # followed by punctuation at a non-word-break. Close the quotes
    # by brute force:
    s = single_quote_start_re.sub("\x04", s)
    s = double_quote_start_re.sub("\x02", s)

    # Special case for double sets of quotes, e.g.:
    #   <p>He said, "'Quoted' words in a larger quote."</p>
    s = double_quote_sets_re.sub("\x01\x03", s)
    s = single_quote_sets_re.sub("\x03\x01", s)

    # Special case for decade abbreviations (the '80s):
    s = decade_abbr_re.sub("\x04", s)

    s = opening_single_quotes_regex.sub("\\1\x03", s)
    s = closing_single_quotes_regex.sub("\\1\x04", s)
    s = closing_single_quotes_regex_2.sub("\\1\x04\\2", s)

    # Any remaining single quotes should be opening ones:
    s = s.replace("'", "\x03")

    s = opening_double_quotes_regex.sub("\\1\x01", s)
    s = closing_double_quotes_regex.sub("\x02", s)
    s = closing_double_quotes_regex_2.sub("\\1\x02", s)

    # Any remaining quotes should be opening ones.
    s = s.replace('"', "\x01")

    # Finally, replace all helpers with quotes.
    return s.replace("\x01", dquotes[0]).replace("\x02", dquotes[1]).\
        replace("\x03", "`").replace("\x04", "'")


BaseTranslator = sphinx.writers.latex.LaTeXTranslator

class DocTranslator(BaseTranslator):

    def __init__(self, *args, **kwargs):
        BaseTranslator.__init__(self, *args, **kwargs)
        self.verbatim = None
    def visit_caption(self, node):
        caption_idx = node.parent.index(node)
        if caption_idx > 0:
            look_node = node.parent.children[caption_idx - 1]
        else:
            look_node = node.parent

        short_caption = unicode(look_node.get('alt', '')).translate(tex_escape_map)
        if short_caption != "":
            short_caption = '[%s]' % short_caption

        self.in_caption += 1
        self.body.append('\\caption%s{' % short_caption)
    def depart_caption(self, node):
        self.body.append('}')
        self.in_caption -= 1

    def visit_Text(self, node):
        if self.verbatim is not None:
            self.verbatim += node.astext()
        else:
            text = self.encode(node.astext())
            if '\\textasciitilde{}' in text:
                text = text.replace('\\textasciitilde{}', '~')
            if not self.no_contractions:
                # text = smartquotes(text)
                text = educate_quotes_latex(text)
            self.body.append(text)

    def visit_table(self, node):
        if self.table:
            raise UnsupportedError(
                '%s:%s: nested tables are not yet implemented.' %
                (self.curfilestack[-1], node.line or ''))

        self.table = sphinx.writers.latex.Table()
        self.table.longtable = False
        self.tablebody = []
        self.tableheaders = []

        # Redirect body output until table is finished.
        self._body = self.body
        self.body = self.tablebody

    def depart_table(self, node):
        self.body = self._body
        colspec = self.table.colspec or ''

        if 'p{' in colspec or 'm{' in colspec or 'b{' in colspec:
            self.body.append('\n\\bodyspacing\n')

        self.body.append('\n\\begin{tabular}')

        if colspec:
            self.body.append(colspec)
        else:
            self.body.append('{|' + ('l|' * self.table.colcount) + '}')

        self.body.append('\n')

        if self.table.caption is not None:
            for id in self.next_table_ids:
                self.body.append(self.hypertarget(id, anchor=False))
            self.next_table_ids.clear()

        self.body.append('\\toprule\n')

        self.body.extend(self.tableheaders)

        self.body.append('\\midrule\n')

        self.body.extend(self.tablebody)

        self.body.append('\\bottomrule\n')

        self.body.append('\n\\end{tabular}\n')

        self.table = None
        self.tablebody = None

    def depart_row(self, node):
        if self.previous_spanning_row == 1:
            self.previous_spanning_row = 0
            self.body.append('\\\\\n')
        else:
            self.body.append('\\\\\n')
        self.table.rowcount += 1

    def depart_literal_block(self, node):
        code = self.verbatim.rstrip('\n')
        lang = self.hlsettingstack[-1][0]
        linenos = code.count('\n') >= self.hlsettingstack[-1][1] - 1
        highlight_args = node.get('highlight_args', {})
        if 'language' in node:
            # code-block directives
            lang = node['language']
            highlight_args['force'] = True
        if 'linenos' in node:
            linenos = node['linenos']
        def warner(msg):
            self.builder.warn(msg, (self.curfilestack[-1], node.line))
        hlcode = self.highlighter.highlight_block(code, lang, warn=warner,
                linenos=linenos, **highlight_args)
        hlcode = hlcode.replace('\$', '$')
        hlcode = hlcode.replace('\%', '%')
        # workaround for Unicode issue
        hlcode = hlcode.replace(u'€', u'@texteuro[]')
        # must use original Verbatim environment and "tabular" environment
        if self.table:
            hlcode = hlcode.replace('\\begin{Verbatim}',
                                    '\\begin{OriginalVerbatim}')
            self.table.has_problematic = True
            self.table.has_verbatim = True
        # get consistent trailer
        hlcode = hlcode.rstrip()[:-14] # strip \end{Verbatim}
        hlcode = hlcode.rstrip() + '\n'
        hlcode = '\n' + hlcode + '\\end{%sVerbatim}\n' % (self.table and 'Original' or '')
        hlcode = hlcode.replace('Verbatim', 'lstlisting')
        begin_bracket = hlcode.find('[')
        end_bracket = hlcode.find(']')
        hlcode = hlcode[:begin_bracket] + '[]' + hlcode[end_bracket+1:]
        self.body.append(hlcode)
        self.verbatim = None

    def visit_figure(self, node):
        ids = ''
        for id in self.next_figure_ids:
            ids += self.hypertarget(id, anchor=False)
        self.next_figure_ids.clear()
        if 'width' in node and node.get('align', '') in ('left', 'right'):
            self.body.append('\\begin{wrapfigure}{%s}{%s}\n\\centering' %
                             (node['align'] == 'right' and 'r' or 'l',
                              node['width']))
            self.context.append(ids + '\\end{wrapfigure}\n')
        else:
            if (not 'align' in node.attributes or
                node.attributes['align'] == 'center'):
                # centering does not add vertical space like center.
                align = '\n\\centering'
                align_end = ''
            else:
                # TODO non vertical space for other alignments.
                align = '\\begin{flush%s}' % node.attributes['align']
                align_end = '\\end{flush%s}' % node.attributes['align']
            self.body.append('\\begin{figure}[tbp]%s\n' % align)
            if any(isinstance(child, nodes.caption) for child in node):
                self.body.append('\\capstart\n')
            self.context.append(ids + align_end + '\\end{figure}\n')

# sphinx.writers.latex.LaTeXTranslator = DocTranslator

class CustomLaTeXTranslator(DocTranslator):
    def astext(self):
            return (#HEADER % self.elements +
                    #self.highlighter.get_stylesheet() +
                    u''.join(self.body)
                    #'\n' + self.elements['footer'] + '\n' +
                    #self.generate_indices() +
                    #FOOTER % self.elements
                    )

    def unknown_departure(self, node):
        if node.tagname == 'only':
            return
        return super(CustomLaTeXTranslator, self).unknown_departure(node)

    def unknown_visit(self, node):
        if node.tagname == 'only':
            return
        return super(CustomLaTeXTranslator, self).unknown_visit(node)

class CustomLaTeXBuilder(sphinx.builders.latex.LaTeXBuilder):

    def write(self, *ignored):
        super(CustomLaTeXBuilder, self).write(*ignored)

        backup_translator = sphinx.writers.latex.LaTeXTranslator
        sphinx.writers.latex.LaTeXTranslator = CustomLaTeXTranslator
        backup_doc = sphinx.writers.latex.BEGIN_DOC
        sphinx.writers.latex.BEGIN_DOC = ''

        # output these as include files
        for docname in ['abstract', 'dedication', 'acknowledgements']:
            destination = FileOutput(
                    destination_path=os.path.join(self.outdir, '%s.inc' % docname),
                    encoding='utf-8')

            docwriter = LaTeXWriter(self)
            doctree = self.env.get_doctree(docname)

            docsettings = OptionParser(
                defaults=self.env.settings,
                components=(docwriter,)).get_default_values()
            doctree.settings = docsettings
            docwriter.write(doctree, destination)

        sphinx.writers.latex.LaTeXTranslator = backup_translator
        sphinx.writers.latex.BEGIN_DOC = backup_doc

    def finish(self, *args, **kwargs):
        super(CustomLaTeXBuilder, self).finish(*args, **kwargs)
        # copy additional files again *after* tex support files so we can override them!
        if self.config.latex_additional_files:
            self.info(bold('copying additional files again...'), nonl=1)
            for filename in self.config.latex_additional_files:
                self.info(' '+filename, nonl=1)
                copyfile(os.path.join(self.confdir, filename),
                         os.path.join(self.outdir, os.path.basename(filename)))
            self.info()

# monkey patch the shit out of it
# sphinx.builders.latex.LaTeXBuilder = CustomLaTeXBuilder



"""
Fixes the table of contents in singlehtml mode so that section titles have the
correct section number in front.
"""

from docutils import nodes
from sphinx import addnodes

def stringize_secnum(secnum):
    return '.'.join(map(str, secnum))

def doctree_resolved(app, doctree, fromdocname):
    if app.builder.name == 'singlehtml':
        secnums = app.builder.env.toc_secnumbers
        for filenode in doctree.traverse(addnodes.start_of_file):
            docname = filenode['docname']
            if docname not in secnums:
                continue

            doc_secnums = secnums[docname]
            first_title_node = filenode.next_node(nodes.title)
            if first_title_node is not None and '' in doc_secnums:
                file_secnum = stringize_secnum(doc_secnums[''])
                title_text_node = first_title_node.children[0]
                newtext = file_secnum + ' ' + title_text_node.astext()
                first_title_node.replace(title_text_node, nodes.Text(newtext))

            for section_node in filenode.traverse(nodes.section):
                for id in section_node['ids']:
                    if '#' + id in doc_secnums:
                        subsection_num = stringize_secnum(doc_secnums['#' + id])
                        first_title_node = section_node.next_node(nodes.title)
                        if first_title_node is not None:
                            title_text_node = first_title_node.children[0]
                            newtext = subsection_num + ' ' + title_text_node.astext()
                            first_title_node.replace(title_text_node, nodes.Text(newtext))

def setup(app):
    app.connect('doctree-resolved', doctree_resolved)



#!/usr/bin/env python
""" Larch Tests Version 1 """
import unittest
import time
import ast
import numpy as np
from sys import version_info

from utils import TestCase

class Test1(TestCase):
    '''testing of asteval'''
    def test1_import(self):
        imported = False
        try:
            import larch
            imported = True
        except:
            pass
        assert(imported)

    def test2_create_interp(self):
        import larch
        self.larch = larch.Interpreter()
        self.symtable = self.larch.symtable

        assert(self.larch is not None)
        assert(self.symtable is not None)

    def test3_set_symbol(self):
        st = self.symtable
        st.set_symbol('_main.a', value = 1)
        st.new_group('g1', s='a string', en=722)
        
        st.set_symbol('_sys.var', value=1)
        st.set_symbol('_main.b',  value='value of b')
        st.set_symbol('_main.g1.t', value='another string')
        st.set_symbol('_main.g1.d', value={'yes': 1, 'no': 0})
        
        st.new_group('g2')
        st.set_symbol('g2.data',value=[1,2,3])
        st.set_symbol('g2.a' ,value = 'hello')
        
        a = st.get_symbol('_main.a')
        assert(a == 1)
        a = st.get_symbol('_main.g1.d')
        assert(a['yes'] == 1)

        a = st.get_symbol('g2.data')
        assert(len(a) == 3)


if __name__ == '__main__':  # pragma: no cover
    for suite in (Test1,):
        suite = unittest.TestLoader().loadTestsFromTestCase(suite)
        unittest.TextTestRunner(verbosity=2).run(suite)



import re
from typing import List, Optional, Iterator
from silx.io.specfile import SpecFile
from . import base


class SpecSingleXasDataSource(base.XasDataSource):
    """SPEC file. Each scan contains 1 XAS spectrum."""

    TYPE = "SPEC"

    def __init__(self, *args, title_regex_pattern: Optional[str] = None, **kw) -> None:
        self.__specfile = None
        if title_regex_pattern:
            title_regex_pattern = re.compile(title_regex_pattern)
        self._title_regex_pattern = title_regex_pattern
        super().__init__(*args, **kw)

    @property
    def _specfile(self):
        if self.__specfile is None:
            self.__specfile = SpecFile(self._filename)
        return self.__specfile

    def get_source_info(self) -> str:
        return f"SPEC: {self._filename}"

    def get_scan(self, scan_name: str) -> Optional[base.XasScan]:
        scan = self._specfile[scan_name]
        description = "\n".join(scan.header)
        return base.XasScan(
            name=scan_name,
            description=description,
            start_time="TODO",
            info=self.get_source_info(),
            labels=scan.labels,
            data=scan.data,
        )

    def get_scan_names(self) -> List[str]:
        return list(self._iter_scan_names())

    def _iter_scan_names(self) -> Iterator[str]:
        for scan in self._specfile:
            if self._title_regex_pattern is not None:
                title = scan.scan_header_dict["S"]
                title = "".join(title.split(" ")[1:])
                if not self._title_regex_pattern.match(title):
                    continue
            yield f"{scan.number}.{scan.order}"



#!/usr/bin/env python
"""
 json utilities for larch objects
"""
import json
import io
import numpy as np
import h5py
from datetime import datetime
from collections import namedtuple
from pathlib import Path, PosixPath
from types import ModuleType
import importlib
import logging


from lmfit import Parameter, Parameters
from lmfit.model import Model, ModelResult
from lmfit.minimizer import Minimizer, MinimizerResult
from lmfit.parameter import SCIPY_FUNCTIONS

from larch import Group, isgroup
from larch.utils.logging  import getLogger
from larch.utils.logging  import _levels as LoggingLevels

HAS_STATE = {}
LarchGroupTypes = {}

def setup_larchtypes():
    global HAS_STATE,  LarchGroupTypes
    if len(HAS_STATE) == 0 or len(LarchGroupTypes)==0:
        try:
            from sklearn.cross_decomposition import PLSRegression
            from sklearn.linear_model import LassoLarsCV, LassoLars, Lasso
            HAS_STATE.update({'PLSRegression': PLSRegression,
                              'LassoLarsCV':LassoLarsCV,
                              'LassoLars': LassoLars, 'Lasso': Lasso})

        except ImportError:
            pass

        from larch import Journal, Group

        HAS_STATE['Journal'] = Journal

        from larch.xafs.feffutils import FeffCalcResults
        HAS_STATE['FeffCalcResults'] = FeffCalcResults

        from larch.xafs import FeffDatFile, FeffPathGroup
        HAS_STATE['FeffDatFile'] = FeffDatFile
        HAS_STATE['FeffPathGroup'] = FeffPathGroup

        from larch import ParameterGroup
        from larch.io.athena_project import AthenaGroup
        from larch.xafs import FeffitDataSet, TransformGroup

        LarchGroupTypes = {'Group': Group,
                           'AthenaGroup': AthenaGroup,
                           'ParameterGroup': ParameterGroup,
                           'FeffitDataSet': FeffitDataSet,
                           'TransformGroup': TransformGroup,
                           'MinimizerResult': MinimizerResult,
                           'Minimizer': Minimizer,
                           'FeffDatFile':  FeffDatFile,
                           'FeffPathGroup': FeffPathGroup,
                           }


def unpack_minimizer(out):
    "this will unpack a minimizer, which can appear in a few places"
    params = out.pop('params')
    userfunc = out.pop('userfcn', None)
    kws = out.pop('kws', None)
    if kws is None:
        kws = {}
    if 'kw' in out:
        kwx = out.pop('kw', None)
        if isinstance(kwx, dict):
            kws.update(kwx)
    for kname in ('scale_covar', 'max_nfev', 'nan_policy'):
        kws[kname] = out.pop(kname)
    mini = Minimizer(userfunc, params, **kws)
    for kname in ('success', 'nfev', 'nfree', 'ndata', 'ier',
                  'errorbars', 'message', 'lmdif_message', 'chisqr',
                  'redchi', 'covar', 'userkws', 'userargs', 'result'):
        setattr(mini, kname, out.pop(kname))
    return mini


def encode4js(obj):
    """return an object ready for json encoding.
    has special handling for many Python types
      numpy array
      complex numbers
      Larch Groups
      Larch Parameters
    """
    setup_larchtypes()
    if obj is None:
        return None
    if isinstance(obj, np.ndarray):
        out = {'__class__': 'Array', '__shape__': obj.shape,
               '__dtype__': obj.dtype.name}
        out['value'] = obj.flatten().tolist()

        if 'complex' in obj.dtype.name:
            out['value'] = [(obj.real).tolist(), (obj.imag).tolist()]
        elif obj.dtype.name == 'object':
            out['value'] = [encode4js(i) for i in out['value']]
        return out
    elif isinstance(obj, (bool, np.bool_)):
        return bool(obj)
    elif isinstance(obj, (int, np.int64, np.int32)):
        return int(obj)
    elif isinstance(obj, (float, np.float64, np.float32)):
        return float(obj)
    elif isinstance(obj, str):
        return str(obj)
    elif isinstance(obj, bytes):
        return obj.decode('utf-8')
    elif isinstance(obj, datetime):
        return {'__class__': 'Datetime', 'isotime': obj.isoformat()}
    elif isinstance(obj, (Path, PosixPath)):
        return {'__class__': 'Path', 'value': obj.as_posix()}
    elif isinstance(obj,(complex, np.complex128)):
        return {'__class__': 'Complex', 'value': (obj.real, obj.imag)}
    elif isinstance(obj, io.IOBase):
        out ={'__class__':  'IOBase', 'class': obj.__class__.__name__,
              'name': obj.name, 'closed': obj.closed,
              'readable': obj.readable(), 'writable': False}
        try:
            out['writable'] = obj.writable()
        except ValueError:
            out['writable'] = False
    elif isinstance(obj, h5py.File):
        return {'__class__': 'HDF5File',
                'value': (obj.name, obj.filename, obj.mode, obj.libver),
                'keys': list(obj.keys())}
    elif isinstance(obj, h5py.Group):
        return {'__class__': 'HDF5Group', 'value': (obj.name, obj.file.filename),
                'keys': list(obj.keys())}
    elif isinstance(obj, slice):
        return {'__class__': 'Slice', 'value': (obj.start, obj.stop, obj.step)}

    elif isinstance(obj, list):
        return {'__class__': 'List', 'value': [encode4js(item) for item in obj]}
    elif isinstance(obj, tuple):
        if hasattr(obj, '_fields'):  # named tuple!
            return {'__class__': 'NamedTuple',
                    '__name__': obj.__class__.__name__,
                    '_fields': obj._fields,
                    'value': [encode4js(item) for item in obj]}
        else:
            return {'__class__': 'Tuple', 'value': [encode4js(item) for item in obj]}
    elif isinstance(obj, dict):
        out = {'__class__': 'Dict'}
        for key, val in obj.items():
            out[encode4js(key)] = encode4js(val)
        return out
    elif isinstance(obj, logging.Logger):
        level = 'DEBUG'
        for key, val in LoggingLevels.items():
            if obj.level == val:
                level = key
        return {'__class__':  'Logger', 'name': obj.name, 'level': level}
    elif isinstance(obj, Model):
        return {'__class__': 'Model', 'value': obj.dumps()}
    elif isinstance(obj, ModelResult):
        return {'__class__': 'ModelResult', 'value': obj.dumps()}

    elif isinstance(obj, Minimizer) and not isinstance(obj, ModelResult):
        out = {'__class__': 'Minimizer'}

        for attr in ('userfcn', 'params', 'kw', 'scale_covar', 'max_nfev',
                     'nan_policy', 'success', 'nfev', 'nfree', 'ndata', 'ier',
                     'errorbars', 'message', 'lmdif_message', 'chisqr',
                     'redchi', 'covar', 'userkws', 'userargs', 'result'):
            out[attr] = encode4js(getattr(obj, attr, None))
        return out
    elif isinstance(obj, MinimizerResult):
        out = {'__class__': 'MinimizerResult'}
        for attr in ('aborted', 'aic', 'bic', 'call_kws', 'chisqr',
                     'covar', 'errorbars', 'ier', 'init_vals',
                     'init_values', 'last_internal_values',
                     'lmdif_message', 'message', 'method', 'ndata', 'nfev',
                     'nfree', 'nvarys', 'params', 'redchi', 'residual',
                     'success', 'var_names'):
            out[attr] = encode4js(getattr(obj, attr, None))
        return out
    elif isinstance(obj, Parameters):
        out = {'__class__': 'Parameters'}
        o_ast = obj._asteval
        out['unique_symbols'] = {key: encode4js(o_ast.symtable[key])
                                 for key in o_ast.user_defined_symbols()}
        out['params'] = [(p.name, p.__getstate__()) for p in obj.values()]
        return out
    elif isinstance(obj, Parameter):
        return {'__class__': 'Parameter', 'name': obj.name, 'state': obj.__getstate__()}
    elif isgroup(obj):
        try:
            classname = obj.__class__.__name__
        except:
            classname = 'Group'
        out = {'__class__': classname}

        if classname == 'ParameterGroup':  # save in order of parameter names
            parnames = dir(obj)
            for par in obj.__params__.keys():
                if par in parnames:
                    out[par] = encode4js(getattr(obj, par))
        else:
            for item in dir(obj):
                out[item] = encode4js(getattr(obj, item))
        return out

    elif isinstance(obj, ModuleType):
        return {'__class__': 'Module',  'value': obj.__name__}
    elif hasattr(obj, '__getstate__') and not callable(obj):
        return {'__class__': 'StatefulObject',
                '__type__': obj.__class__.__name__,
                'value': encode4js(obj.__getstate__())}
    elif isinstance(obj, type):
        return {'__class__': 'Type',  'value': repr(obj),
                'module': getattr(obj, '__module__', None)}
    elif callable(obj):
        return {'__class__': 'Method', '__name__': repr(obj)}
    elif hasattr(obj, 'dumps'):
        print("Encode Warning: using dumps for ", obj)
        return {'__class__': 'DumpableObject', 'value': obj.dumps()}
    else:
        print("Encode Warning: generic object dump for ", repr(obj))
        out = {'__class__': 'Object', '__repr__': repr(obj),
               '__classname__': obj.__class__.__name__}
        for attr in dir(obj):
            if attr.startswith('__') and attr.endswith('__'):
                continue
            thing = getattr(obj, attr)
            if not callable(thing):
                # print("will try to encode thing ", thing, type(thing))
                out[attr] = encode4js(thing)
        return out

    return obj

def decode4js(obj):
    """
    return decoded Python object from encoded object.

    """
    if not isinstance(obj, dict):
        return obj
    setup_larchtypes()
    out = obj
    classname = obj.pop('__class__', None)
    if classname is None:
        return obj

    if classname == 'Complex':
        out = obj['value'][0] + 1j*obj['value'][1]
    elif classname in ('List', 'Tuple', 'NamedTuple'):
        out = []
        for item in obj['value']:
            out.append(decode4js(item))
        if classname == 'Tuple':
            out = tuple(out)
        elif classname == 'NamedTuple':
            out = namedtuple(obj['__name__'], obj['_fields'])(*out)
    elif classname == 'Array':
        if obj['__dtype__'].startswith('complex'):
            re = np.asarray(obj['value'][0], dtype='double')
            im = np.asarray(obj['value'][1], dtype='double')
            out = re + 1j*im
        elif obj['__dtype__'].startswith('object'):
            val = [decode4js(v) for v in obj['value']]
            out = np.array(val,  dtype=obj['__dtype__'])

        else:
            out = np.asarray(obj['value'], dtype=obj['__dtype__'])
        out.shape = obj['__shape__']
    elif classname in ('Dict', 'dict'):
        out = {}
        for key, val in obj.items():
            out[key] = decode4js(val)
    elif classname == 'Datetime':
        obj = datetime.fromisoformat(obj['isotime'])
    elif classname in ('Path', 'PosixPath'):
        obj = Path(obj['value'])

    elif classname == 'IOBase':
        mode = 'r'
        if obj['readable']  and obj['writable']:
            mode = 'a'
        elif not obj['readable']  and obj['writable']:
            mode = 'w'
        out = open(obj['name'], mode=mode)
        if obj['closed']:
            out.close()

    elif classname == 'Module':
        out = importlib.import_module(obj.__name__)

    elif classname == 'Parameters':
        out = Parameters()
        out.clear()
        unique_symbols = {key: decode4js(obj['unique_symbols'][key]) for key
                          in obj['unique_symbols']}

        state = {'unique_symbols': unique_symbols, 'params': []}
        for name, parstate in obj['params']:
            par = Parameter(decode4js(name))
            par.__setstate__(decode4js(parstate))
            state['params'].append(par)
        out.__setstate__(state)
    elif classname in ('Parameter', 'parameter'):
        name = decode4js(obj['name'])
        state = decode4js(obj['state'])
        out = Parameter(name)
        out.__setstate__(state)

    elif classname == 'Model':
        mod = Model(lambda x: x)
        out = mod.loads(decode4js(obj['value']))

    elif classname == 'ModelResult':
        params = Parameters()
        res = ModelResult(Model(lambda x: x, None), params)
        out = res.loads(decode4js(obj['value']))

    elif classname == 'Logger':
        out = getLogger(obj['name'], level=obj['level'])

    elif classname == 'StatefulObject':
        dtype = obj.get('__type__')
        if dtype in HAS_STATE:
            out = HAS_STATE[dtype]()
            out.__setstate__(decode4js(obj.get('value')))
        elif dtype == 'Minimizer':
            out = unpack_minimizer(out['value'])
        else:
            print(f"Warning: cannot re-create stateful object of type '{dtype}'")

    elif classname in LarchGroupTypes:
        out = {}
        for key, val in obj.items():
            if (isinstance(val, dict) and
                val.get('__class__', None) == 'Method' and
                val.get('__name__', None) is not None):
                pass  # ignore class methods for subclassed Groups
            else:
                out[key] = decode4js(val)
        if classname == 'Minimizer':
            out = unpack_minimizer(out)
        elif classname == 'FeffDatFile':
            from larch.xafs import FeffDatFile
            path = FeffDatFile()
            path._set_from_dict(**out)
            out = path
        else:
            out = LarchGroupTypes[classname](**out)
    elif classname == 'Method':
        mname = obj.get('__name__', '')
        if 'ufunc' in mname:
            mname = mname.replace('<ufunc', '').replace('>', '').replace("'","").strip()
        out = SCIPY_FUNCTIONS.get(mname, None)

    else:
        print("cannot decode ", classname, repr(obj)[:100])
    return out



#!/usr/bin/env python
"""
  XAFS pre-edge subtraction, normalization algorithms
"""
import numpy as np

from lmfit import Parameters, Minimizer, report_fit
from xraydb import guess_edge
from larch import Group, Make_CallArgs, parse_group_args

from larch.math import (index_of, index_nearest, interp, smooth,
                        polyfit, remove_dups, remove_nans, remove_nans2)
from .xafsutils import set_xafsGroup, TINY_ENERGY

MODNAME = '_xafs'
MAX_NNORM = 5

@Make_CallArgs(["energy","mu"])
def find_e0(energy, mu=None, group=None, _larch=None):
    """calculate E0, the energy threshold of absorption, or 'edge energy', given mu(E).

    E0 is found as the point with maximum derivative with some checks to avoid spurious glitches.

    Arguments:
        energy (ndarray or group): array of x-ray energies, in eV, or group
        mu     (ndaarray or None): array of mu(E) values
        group  (group or None):    output group
        _larch (larch instance or None):  current larch session.

    Returns:
        float: Value of e0. If a group is provided, group.e0 will also be set.

    Notes:
        1. Supports :ref:`First Argument Group` convention, requiring group members `energy` and `mu`
        2. Supports :ref:`Set XAFS Group` convention within Larch or if `_larch` is set.
    """
    energy, mu, group = parse_group_args(energy, members=('energy', 'mu'),
                                         defaults=(mu,), group=group,
                                         fcn_name='find_e0')
    # first find e0 without smoothing, then refine with smoothing
    e1, ie0, estep1 = _finde0(energy, mu, estep=None, use_smooth=False)
    istart = max(3, ie0-75)
    istop  = min(ie0+75, len(energy)-3)
    # sanity check: e0 should not be in first 5% of energy point: avoids common glitches
    if ie0 < 0.05*len(energy):
        e1 = energy.mean()
        istart = max(3, ie0-20)
        istop = len(energy)-3

    # for the smoothing energy, we use and energy step that is an average of
    # the observed minimimum energy step (which could be ridiculously low)
    # and a scaled value of the initial e0 (0.2 eV and 5000 eV, 0.4 eV at 10000 eV)
    # print("Find E0 step 1 ", e1, ie0, len(energy), estep1, istart, istop)
    estep = 0.5*(max(0.01, min(1.0, estep1)) + max(0.01, min(1.0, e1/25000.)))
    e0, ix, ex = _finde0(energy[istart:istop], mu[istart:istop], estep=estep, use_smooth=True)
    if ix < 1 :
        e0 = energy[istart+2]
    if group is not None:
        group = set_xafsGroup(group, _larch=_larch)
        group.e0 = e0
    return e0

def find_energy_step(energy, frac_ignore=0.01, nave=10):
    """robustly find energy step in XAS energy array,
    ignoring the smallest fraction of energy steps (frac_ignore),
    and averaging over the next `nave` values
    """
    nskip = int(frac_ignore*len(energy))
    e_ordered = np.where(np.diff(np.argsort(energy))==1)[0]  # where energy step are in order
    ediff = np.diff(energy[e_ordered][nskip:-nskip])
    return ediff[np.argsort(ediff)][nskip:nskip+nave].mean()


def _finde0(energy, mu_input, estep=None, use_smooth=True):
    "internally used by find e0 "

    en = remove_dups(energy, tiny=TINY_ENERGY)
    ordered = np.where(np.diff(np.argsort(en))==1)[0]
    en = en[ordered]
    mu = mu_input[ordered]
    if len(en.shape) > 1:
        en = en.squeeze()
    if len(mu.shape) > 1:
        mu = mu.squeeze()
    if estep is None:
        estep = find_energy_step(en)


    nmin = max(3, int(len(en)*0.02))
    if use_smooth:
        dmu = smooth(en, np.gradient(mu)/np.gradient(en), xstep=estep, sigma=estep)
    else:
        dmu = np.gradient(mu)/np.gradient(en)
    # find points of high derivative
    dmu[np.where(~np.isfinite(dmu))] = -1.0
    dm_min = dmu[nmin:-nmin].min()
    dm_ptp = max(1.e-10, np.ptp(dmu[nmin:-nmin]))
    dmu = (dmu - dm_min)/dm_ptp

    dhigh = 0.60 if len(en) > 20 else 0.30
    high_deriv_pts = np.where(dmu > dhigh)[0]
    if len(high_deriv_pts) < 3:
        for _ in range(2):
            if len(high_deriv_pts) > 3:
                break
            dhigh *= 0.5
            high_deriv_pts = np.where(dmu > dhigh)[0]

    if len(high_deriv_pts) < 3:
        high_deriv_pts = np.where(np.isfinite(dmu))[0]

    imax, dmax = 0, 0
    for i in high_deriv_pts:
        if i < nmin or i > len(en) - nmin:
            continue
        if (dmu[i] > dmax and
            (i+1 in high_deriv_pts) and
            (i-1 in high_deriv_pts)):
            imax, dmax = i, dmu[i]
    return en[imax], imax, estep

def flat_resid(pars, en, mu):
    return pars['c0'] + en * (pars['c1'] + en * pars['c2']) - mu

def preedge(energy, mu, e0=None, step=None, nnorm=None, nvict=0, pre1=None,
            pre2=None, norm1=None, norm2=None):
    """pre edge subtraction, normalization for XAFS (straight python)

    This performs a number of steps:
       1. determine E0 (if not supplied) from max of deriv(mu)
       2. fit a line to the region below the edge
       3. fit a polymonial to the region above the edge
       4. extrapolate the two curves to E0 and take their difference
          to determine the edge jump

    Arguments
    ----------
    energy:  array of x-ray energies, in eV
    mu:      array of mu(E)
    e0:      edge energy, in eV.  If None, it will be determined here.
    step:    edge jump.  If None, it will be determined here.
    pre1:    low E range (relative to E0) for pre-edge fit
    pre2:    high E range (relative to E0) for pre-edge fit
    nvict:   energy exponent to use for pre-edg fit.  See Note
    norm1:   low E range (relative to E0) for post-edge fit
    norm2:   high E range (relative to E0) for post-edge fit
    nnorm:   degree of polynomial (ie, nnorm+1 coefficients will be found) for
             post-edge normalization curve. Default=None -- see note.
    Returns
    -------
      dictionary with elements (among others)
          e0          energy origin in eV
          edge_step   edge step
          norm        normalized mu(E)
          pre_edge    determined pre-edge curve
          post_edge   determined post-edge, normalization curve

    Notes
    -----
    1  pre_edge: a line is fit to mu(energy)*energy**nvict over the region,
       energy=[e0+pre1, e0+pre2]. pre1 and pre2 default to None, which will set
           pre1 = e0 - 2nd energy point, rounded to 5 eV
           pre2 = roughly pre1/3.0, rounded to 5 eV

    2  post-edge: a polynomial of order nnorm is fit to mu(energy)*energy**nvict
       between energy=[e0+norm1, e0+norm2]. nnorm, norm1, norm2 default to None,
       which will set:
         nnorm = 2 in norm2-norm1>300, 1 if norm2-norm1>30, or 0 if less.
         norm2 = max energy - e0, rounded to 5 eV
         norm1 = roughly min(150, norm2/3.0), rounded to 5 eV
    """

    energy, mu = remove_nans2(energy, mu)
    energy = remove_dups(energy, tiny=TINY_ENERGY)
    if energy.size <= 1:
        raise ValueError("energy array must have at least 2 points")
    if e0 is None or e0 < energy[1] or e0 > energy[-2]:
        e0 = find_e0(energy, mu)
    ie0 = index_nearest(energy, e0)
    e0 = energy[ie0]

    if pre1 is None:
        # skip first energy point, often bad
        if ie0 > 20:
            pre1  = 5.0*round((energy[1] - e0)/5.0)
        else:
            pre1  = 2.0*round((energy[1] - e0)/2.0)
    pre1 = max(pre1,  (min(energy) - e0))
    if pre2 is None:
        pre2 = 0.5*pre1
    if pre1 > pre2:
        pre1, pre2 = pre2, pre1
    ipre1 = index_of(energy-e0, pre1)
    ipre2 = index_of(energy-e0, pre2)
    if ipre2 < ipre1 + 2 + nvict:
        pre2 = (energy-e0)[int(ipre1 + 2 + nvict)]

    if norm2 is None:
        norm2 = 5.0*round((max(energy) - e0)/5.0)
    if norm2 < 0:
        norm2 = max(energy) - e0 - norm2
    norm2 = min(norm2, (max(energy) - e0))
    if norm1 is None:
        norm1 = min(25, 5.0*round(norm2/15.0))

    if norm1 > norm2:
        norm1, norm2 = norm2, norm1

    norm1 = min(norm1, norm2 - 10)
    if nnorm is None:
        nnorm = 2
        if norm2-norm1 < 300: nnorm = 1
        if norm2-norm1 <  30: nnorm = 0
    nnorm = max(min(nnorm, MAX_NNORM), 0)
    # preedge
    p1 = index_of(energy, pre1+e0)
    p2 = index_nearest(energy, pre2+e0)
    if p2-p1 < 2:
        p2 = min(len(energy), p1 + 2)

    omu  = mu*energy**nvict
    ex = remove_nans(energy[p1:p2], interp=True)
    mx = remove_nans(omu[p1:p2], interp=True)

    precoefs = polyfit(ex, mx, 1)
    pre_edge = (precoefs[0] + energy*precoefs[1]) * energy**(-nvict)
    # normalization
    p1 = index_of(energy, norm1+e0)
    p2 = index_nearest(energy, norm2+e0)
    if p2-p1 < 2:
        p2 = min(len(energy), p1 + 2)
    if p2-p1 < 2:
        p1 = p1-2

    presub = (mu-pre_edge)[p1:p2]
    coefs = polyfit(energy[p1:p2], presub, nnorm)
    post_edge = 1.0*pre_edge
    norm_coefs = []
    for n, c in enumerate(coefs):
        post_edge += c * energy**(n)
        norm_coefs.append(c)
    edge_step = step
    if edge_step is None:
        edge_step = post_edge[ie0] - pre_edge[ie0]
    edge_step = max(1.e-12, abs(float(edge_step)))
    norm = (mu - pre_edge)/edge_step
    return {'e0': e0, 'edge_step': edge_step, 'norm': norm,
            'pre_edge': pre_edge, 'post_edge': post_edge,
            'norm_coefs': norm_coefs, 'nvict': nvict,
            'nnorm': nnorm, 'norm1': norm1, 'norm2': norm2,
            'pre1': pre1, 'pre2': pre2, 'precoefs': precoefs}

@Make_CallArgs(["energy","mu"])
def pre_edge(energy, mu=None, group=None, e0=None, step=None, nnorm=None,
             nvict=0, pre1=None, pre2=None, norm1=None, norm2=None,
             make_flat=True, _larch=None):
    """pre edge subtraction, normalization for XAFS

    This performs a number of steps:
       1. determine E0 (if not supplied) from max of deriv(mu)
       2. fit a line of polymonial to the region below the edge
       3. fit a polymonial to the region above the edge
       4. extrapolate the two curves to E0 and take their difference
          to determine the edge jump

    Arguments
    ----------
    energy:  array of x-ray energies, in eV, or group (see note 1)
    mu:      array of mu(E)
    group:   output group
    e0:      edge energy, in eV. If None, it will be determined here.
    step:    edge jump.  If None, it will be determined here.
    pre1:    low E range (relative to E0) for pre-edge fit
    pre2:    high E range (relative to E0) for pre-edge fit
    nvict:   energy exponent to use for pre-edg fit.  See Notes.
    norm1:   low E range (relative to E0) for post-edge fit
    norm2:   high E range (relative to E0) for post-edge fit
    nnorm:   degree of polynomial (ie, nnorm+1 coefficients will be found) for
             post-edge normalization curve. See Notes.
    make_flat: boolean (Default True) to calculate flattened output.

    Returns
    -------
      None: The following attributes will be written to the output group:
        e0          energy origin
        edge_step   edge step
        norm        normalized mu(E), using polynomial
        norm_area   normalized mu(E), using integrated area
        flat        flattened, normalized mu(E)
        pre_edge    determined pre-edge curve
        post_edge   determined post-edge, normalization curve
        dmude       derivative of normalized mu(E)
        d2mude      second derivative of normalized mu(E)

    (if the output group is None, _sys.xafsGroup will be written to)

    Notes
    -----
      1. Supports `First Argument Group` convention, requiring group members `energy` and `mu`.
      2. Support `Set XAFS Group` convention within Larch or if `_larch` is set.
      3. pre_edge: a line is fit to mu(energy)*energy**nvict over the region,
         energy=[e0+pre1, e0+pre2]. pre1 and pre2 default to None, which will set
             pre1 = e0 - 2nd energy point, rounded to 5 eV
             pre2 = roughly pre1/3.0, rounded to 5 eV
      4. post-edge: a polynomial of order nnorm is fit to mu(energy)*energy**nvict
         between energy=[e0+norm1, e0+norm2]. nnorm, norm1, norm2 default to None,
         which will set:
              norm2 = max energy - e0, rounded to 5 eV
              norm1 = roughly min(150, norm2/3.0), rounded to 5 eV
              nnorm = 2 in norm2-norm1>300, 1 if norm2-norm1>30, or 0 if less.
      5. flattening fits a quadratic curve (no matter nnorm) to the post-edge
         normalized mu(E) and subtracts that curve from it.
    """
    energy, mu, group = parse_group_args(energy, members=('energy', 'mu'),
                                         defaults=(mu,), group=group,
                                         fcn_name='pre_edge')
    energy, mu = remove_nans2(energy, mu)
    if len(energy.shape) > 1:
        energy = energy.squeeze()
    if len(mu.shape) > 1:
        mu = mu.squeeze()

    out_of_order = np.where(np.diff(np.argsort(energy))!=1)[0]
    if len(out_of_order) > 0:
        order = np.argsort(energy)
        energy = energy[order]
        mu = mu[order]
    energy = remove_dups(energy, tiny=TINY_ENERGY)

    if group is not None and e0 is None:
        e0 = getattr(group, 'e0', None)
    pre_dat = preedge(energy, mu, e0=e0, step=step, nnorm=nnorm,
                      nvict=nvict, pre1=pre1, pre2=pre2, norm1=norm1,
                      norm2=norm2)
    group = set_xafsGroup(group, _larch=_larch)

    e0    = pre_dat['e0']
    norm  = pre_dat['norm']
    norm1 = pre_dat['norm1']
    norm2 = pre_dat['norm2']
    # generate flattened spectra, by fitting a quadratic to .norm
    # and removing that.

    ie0 = index_nearest(energy, e0)
    p1 = index_of(energy, norm1+e0)
    p2 = index_nearest(energy, norm2+e0)
    if p2-p1 < 2:
        p2 = min(len(energy), p1 + 2)

    group.e0 = e0
    group.norm = norm
    group.flat = 1.0*norm
    group.norm_poly = 1.0*norm

    if make_flat:
        pre_edge = pre_dat['pre_edge']
        post_edge = pre_dat['post_edge']
        edge_step = pre_dat['edge_step']
        flat_residue = (post_edge - pre_edge)/edge_step
        flat = norm - flat_residue + flat_residue[ie0]
        flat[:ie0] = norm[:ie0]
        group.flat = flat

        enx = remove_nans(energy[p1:p2], interp=True)
        mux = remove_nans(norm[p1:p2], interp=True)

        # enx, mux = (energy[p1:p2], norm[p1:p2])
        fpars = Parameters()
        ncoefs = len(pre_dat['norm_coefs'])
        fpars.add('c0', value=1.0, vary=True)
        fpars.add('c1', value=0.0, vary=False)
        fpars.add('c2', value=0.0, vary=False)
        if ncoefs > 1:
            fpars['c1'].set(value=1.e-5, vary=True)
            if ncoefs > 2:
                fpars['c2'].set(value=1.e-5, vary=True)

        try:
            fit = Minimizer(flat_resid, fpars, fcn_args=(enx, mux))
            result = fit.leastsq()
            fc0 = result.params['c0'].value
            fc1 = result.params['c1'].value
            fc2 = result.params['c2'].value

            flat_diff = fc0 + energy * (fc1 + energy * fc2)
            flat_alt  = norm - flat_diff  + flat_diff[ie0]
            flat_alt[:ie0]  = norm[:ie0]
            group.flat_coefs = (fc0, fc1, fc2)
            group.flat_alt = flat_alt
        except:
            pass

    group.dmude = np.gradient(norm)/np.gradient(energy)
    group.d2mude = np.gradient(group.dmude)/np.gradient(energy)
    group.edge_step  = pre_dat['edge_step']
    group.edge_step_poly = pre_dat['edge_step']
    group.pre_edge   = pre_dat['pre_edge']
    group.post_edge  = pre_dat['post_edge']

    group.pre_edge_details = Group()
    for attr in ('pre1', 'pre2', 'norm1', 'norm2', 'nnorm', 'nvict'):
        setattr(group.pre_edge_details, attr, pre_dat.get(attr, None))

    group.pre_edge_details.pre_slope  = pre_dat['precoefs'][1]
    group.pre_edge_details.pre_offset = pre_dat['precoefs'][0]

    for i in range(MAX_NNORM):
        if hasattr(group, 'norm_c%i' % i):
            delattr(group, 'norm_c%i' % i)
    for i, c in enumerate(pre_dat['norm_coefs']):
        setattr(group.pre_edge_details, 'norm_c%i' % i, c)

    # guess element and edge
    group.atsym = getattr(group, 'atsym', None)
    group.edge = getattr(group, 'edge', None)

    if group.atsym is None or group.edge is None:
        _atsym, _edge = guess_edge(group.e0)
        if group.atsym is None: group.atsym = _atsym
        if group.edge is None:  group.edge = _edge
    return

def energy_align(group, reference, array='dmude', emin=-15, emax=35):
    """
    align XAFS data group to a reference group

    Arguments
    ---------
    group      Larch group for spectrum to be aligned (see Note 1)
    reference  Larch group for reference spectrum     (see Note 1)
    array      string of 'dmude', 'norm', or 'mu'     (see Note 2) ['dmude']
    emin       float, min energy relative to e0 of reference for alignment [-15]
    emax       float, max energy relative to e0 of reference for alignment [+35]

    Returns
    -------
    eshift   energy shift to add to group.energy to match reference.
             This value will also be written to group.eshift

    Notes
    -----
      1.  Both group and reference must be XAFS data, with arrays of 'energy' and 'mu'.
          The reference group must already have an e0 value set.

      2.  The alignment can be done with 'mu' or 'dmude'.  If it does not exist, the
          dmude array will be built for group and reference.

    """
    if not (hasattr(group, 'energy') and hasattr(group, 'mu')):
        raise ValueError("group must have attributes 'energy' and 'mu'")

    if not hasattr(group, 'dmude'):
        mu = getattr(group, 'norm', getattr(group, 'mu'))
        en = getattr(group, 'energy')
        group.dmude = gradient(mu)/gradient(en)


    if not (hasattr(reference, 'energy') and hasattr(reference, 'mu')
            and hasattr(reference, 'e0') ):
        raise ValueError("reference must have attributes 'energy', 'mu', and 'e0'")

    if not hasattr(reference, 'dmude'):
        mu = getattr(reference, 'norm', getattr(reference, 'mu'))
        en = getattr(reference, 'energy')
        reference.dmude = gradient(mu)/gradient(en)

    xdat = group.energy[:]*1.0
    xref = reference.energy[:]*1.0
    ydat = group.dmude[:]*1.0
    yref = reference.dmude[:]*1.0
    if array == 'mu':
        ydat = group.mu[:]*1.0
        yref = reference.mu[:]*1.0
    elif array == 'norm':
        ydat = group.norm[:]*1.0
        yref = reference.norm[:]*1.0

    xdat = remove_nans(xdat[:], interp=True)
    ydat = remove_nans(ydat[:], interp=True)
    xref = remove_nans(xref[:], interp=True)
    yref = remove_nans(yref[:], interp=True)

    i1 = index_of(xref, reference.e0-emin)
    i2 = index_of(xref, reference.e0+emax)

    def align_resid(params, xdat, ydat, xref, yref, i1, i2):
        "fit residual"
        newx = xdat + params['eshift'].value
        scale = params['scale'].value
        ytmp = interp(newx, ydat, xref, kind='cubic')
        return (ytmp*scale - yref)[i1:i2]

    params = Parameters()
    params.add('eshift', value=0, min=-50, max=50)
    params.add('scale', value=1, min=0, max=50)

    try:
        fit = Minimizer(align_resid, params,
                        fcn_args=(xdat, ydat, xref, yref, i1, i2))
        result = fit.leastsq()
        eshift = result.params['eshift'].value
    except:
        eshift = 0

    group.eshift = eshift
    return eshift



#!/usr/bin/env python
import sys
from traceback import format_tb
import time
from datetime import datetime
from gzip import GzipFile
import io
import copy
import json
import logging

from charset_normalizer import from_bytes
from .gformat import gformat, getfloat_attr
from .paths import uname, bindir, unixpath, get_homedir, path_split, get_cwd
from .debugtimer import debugtimer

from .strutils import (fixName, isValidName, isNumber, bytes2str, str2bytes,
                       fix_filename, fix_varname, strip_quotes, isLiteralStr,
                       strip_comments, asfloat, find_delims, version_ge,
                       unique_name, get_sessionid, strict_ascii)

from .shellutils import (_more, _parent, ls, cd, cwd, mkdir)

logging.basicConfig(format='%(levelname)s [%(asctime)s]: %(message)s',
                    datefmt='%Y-%m-%d %H:%M:%S', level=logging.WARNING)

def format_exception(with_traceback=True):
    """return exception message as list of strings,
    optionally including traceback
    """
    etype, exc, tb = sys.exc_info()
    out = []
    if with_traceback:
        out = ["Traceback (most recent calls last):"]
        for tline in format_tb(tb):
            if tline.endswith('\n'): tline = tline[:-1]
            out.append(tline)
        out.append('\n')
    out.append(f"{etype.__name__}: {exc}")
    return out


def write_log(msg, level='debug'):
    f = logging.debug
    if level in ('warn', 'warning', logging.WARNING):
        f = logging.warning
    elif level in ('info', logging.INFO):
        f = logging.info
    elif level in ('error', logging.ERROR):
        f = logging.error
    elif level in ('critical', logging.CRITICAL):
        f = logging.critical
    return f(msg)

def log_warning(msg):
    return logging.warning(msg)

def log_debug(msg):
    return logging.debug(msg)

def log_info(msg):
    return logging.info(msg)

def log_error(msg):
    return logging.error(msg)

def log_critical(msg):
    return logging.critical(msg)


def is_gzip(filename):
    "is a file gzipped?"
    with open(unixpath(filename), 'rb') as fh:
        return fh.read(3) == b'\x1f\x8b\x08'
    return False

def read_textfile(filename, size=None):
    """read text from a file as string

    Argument
    --------
    filename  (str or file): name of file to read or file-like object
    size  (int or None): number of bytes to read

    Returns
    -------
    text of file as string.

    Notes
    ------
    1. the encoding is detected with charset_normalizer.from_bytes
       which is then used to decode bytes read from file.
    2. line endings are normalized to be '\n', so that
       splitting on '\n' will give a list of lines.
    3. if filename is given, it can be a gzip-compressed file
    """
    text = ''

    def decode(bytedata):
        return str(from_bytes(bytedata).best())

    if isinstance(filename, io.IOBase):
        text = filename.read(size)
        if filename.mode == 'rb':
            text = decode(text)
    else:
        fopen = GzipFile if is_gzip(filename) else open
        with fopen(unixpath(filename), 'rb') as fh:
            text = decode(fh.read(size))
    return text.replace('\r\n', '\n').replace('\r', '\n')


def group2dict(group, _larch=None):
    "return dictionary of group members"
    return group.__dict__

def dict2group(d, _larch=None):
    "return group created from a dictionary"
    from larch import Group
    return Group(**d)

def copy_group(group, _larch=None):
    from larch import Group
    out = Group(datatype=getattr(group, 'datatype', 'unknown'),
                copied_from=getattr(group, 'groupname', repr(group)))
    for attr in dir(group):
        try:
            setattr(out, attr, copy.deepcopy(getattr(group, attr)))
        except:
            print(f"cannot copy attribute {attr} from {group}")
    return out

def copy_xafs_group(group, _larch=None):
    """specialized group copy for XAFS data groups"""
    from larch import Group
    out = Group(datatype=getattr(group, 'datatype', 'unknown'),
                copied_from=getattr(group, 'groupname', repr(group)))

    for attr in dir(group):
        if attr not in ('norm', 'flat', 'deriv', 'deconv',
                        'post_edge', 'pre_edge', 'norm_mback',
                        'norm_vict', 'norm_poly'):
            try:
                val = copy.deepcopy(getattr(group, attr))
                setattr(out, attr, val)
            except ValueError:
                pass
    return out


def isotime(t=None, with_tzone=False, filename=False):
    if t is None:
        t = time.time()
    sout = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(t))
    if with_tzone:
        sout = "%s-%2.2i:00" % (sout, time.timezone/3600)
    if filename:
        sout = sout.replace(' ', '_').replace(':', '')
    return sout

def time_ago(timestamp, precision=2):
    """
    give a human-readable 'time ago' from the timestamp.

    The output gives day, hours, minutes, seconds:
       52 days, 1 hour

    the `precision` field gives the number of significant time units to
    show.  This defaults to 2:
       'N days, H hours',
       'N hours, M minutes'
    """
    def format(x, unit):
        return "%d %s%s" % (x, unit, "s" if x > 1 else "")

    tdiff = datetime.now() - datetime.fromtimestamp(timestamp)
    days = tdiff.days
    hours = tdiff.seconds//3600
    minutes = tdiff.seconds%3600//60
    seconds = tdiff.seconds%3600%60

    out = []
    if days > 0:
        out.append(format(days, "day"))
    if hours > 0:
        out.append(format(hours, "hour"))
    if minutes > 0:
        out.append(format(minutes, "minute"))
    out.append(format(seconds, "second"))
    return ", ".join(out[:precision])

def json_dump(data, filename):
    """
    dump object or group to file using json
    """
    from .jsonutils import encode4js
    with open(unixpath(filename), 'w') as fh:
        fh.write(json.dumps(encode4js(data)))
        fh.write('\n')

def json_load(filename):
    """
    load object from json dump file
    """
    from .jsonutils import decode4js
    with open(unixpath(filename), 'rb') as fh:
        data = fh.read().decode('utf-8')
    return decode4js(json.loads(data))

def _larch_init(_larch):
    """initialize xrf"""
    from ..symboltable import Group
    _larch.symtable._sys.display = Group(use_color=True,
                                         colors=dict(text={'color': 'black'},
                                                     text2={'color': 'blue'},
                                                     error={'color': 'red'}))

_larch_builtins = dict(copy=copy.copy, deepcopy=copy.deepcopy, more= _more,
                       parent=_parent, ls=ls, mkdir=mkdir, cd=cd,
                       cwd=cwd, group2dict=group2dict,
                       copy_group=copy_group, copy_xafs_group=copy_xafs_group,
                       dict2group=dict2group, debugtimer=debugtimer,
                       isotime=isotime, json_dump=json_dump,
                       json_load=json_load, gformat=gformat,
                       fix_filename=fix_filename,
                       fix_varname=fix_varname,
                       strip_quotes=strip_quotes)



import re
from docutils import nodes
import sphinx.writers.html

BaseTranslator = sphinx.writers.html.HTMLTranslator

class CustomHTMLTranslator(BaseTranslator):
    def visit_tabular_col_spec(self, node):
        self.table_spec = re.split(r'[\s\|]+', node['spec'])
        raise nodes.SkipNode

    def bulk_text_processor(self, text):
        if '~' in text:
            text = text.replace('~', '&nbsp;')
        return text

    def visit_entry(self, node):
        atts = {'class': []}
        if isinstance(node.parent.parent, nodes.thead):
            atts['class'].append('head')
        if node.parent.parent.parent.stubs[node.parent.column]:
            # "stubs" list is an attribute of the tgroup element
            atts['class'].append('stub')
        if atts['class']:
            tagname = 'th'
            atts['class'] = ' '.join(atts['class'])
        else:
            tagname = 'td'
            del atts['class']
        table_spec = getattr(self, 'table_spec', None)
        if (tagname == 'td' or tagname == 'th') and table_spec is not None:
            if len(table_spec) > node.parent.column:
                colspec = table_spec[node.parent.column]

                horiz_align = ''
                vert_align = ''

                if 'raggedright' in colspec or colspec == 'l':
                    horiz_align = ' align-left'
                elif 'raggedleft' in colspec or colspec == 'r':
                    horiz_align = ' align-right'
                elif 'center' in colspec or colspec == 'c':
                    horiz_align = ' align-center'

                if 'p{' in colspec:
                    vert_align = ' align-top'
                elif 'm{' in colspec:
                    vert_align = ' align-middle'
                elif 'b{' in colspec:
                    vert_align = ' align-bottom'

                align_type = {'l': 'left',
                              'r': 'right',
                              'c': 'center'}
                atts['class'] = (atts.get('class', '') + horiz_align + vert_align)
        node.parent.column += 1
        if 'morerows' in node:
            atts['rowspan'] = node['morerows'] + 1
        if 'morecols' in node:
            atts['colspan'] = node['morecols'] + 1
            node.parent.column += node['morecols']
        self.body.append(self.starttag(node, tagname, '', **atts))
        self.context.append('</%s>\n' % tagname.lower())
        if len(node) == 0:              # empty cell
            self.body.append('&nbsp;')
        self.set_first_last(node)

sphinx.writers.html.SmartyPantsHTMLTranslator = CustomHTMLTranslator



#!/usr/bin/env python

import os
import sys
import copy
import os
import time
import gc

import numpy as np
from .. import Group
from ..utils import read_textfile
from ..utils.strutils import bytes2str

from . import XDIFile, XDIFileException

from .xsp3_hdf5 import XSPRESS3_TAU, estimate_icr

def read_gsexdi(fname, nmca=128, bad=None, **kws):
    """Read GSE XDI Scan Data to larch group,
    summing ROI data for MCAs and apply deadtime corrections
    """

    MAX_MCAS = nmca
    group = Group()
    group.__name__ ='GSE XDI Data file %s' % fname
    xdi = XDIFile(str(fname))

    group.path = fname
    path, suffix = os.path.split(fname)
    group.filename = suffix
    group.npts = xdi.npts
    group.bad_channels = bad
    for family in ('scan', 'beamline', 'mono', 'facility'):
        for key, val in xdi.attrs.get(family, {}).items():
            if '||' in val:
                val, addr = val.split('||')
            try:
                val = float(val)
            except:
                pass
            setattr(group, "%s_%s" % (family, key), val)

    group._xdi = Group()

    for thing in ('array_addrs', 'array_labels', 'array_units', 'attrs',
                  'comments', 'status', 'user_labels', 'xdi_libversion',
                  'xdi_version', 'xdi_pyversion'):

        val = copy.deepcopy(getattr(xdi, thing, None))
        if isinstance(val, bytes):
            val = val.decode('utf-8')
        setattr(group._xdi, thing, val)

    scanparams = xdi.attrs.get('scanparameters', None)
    if scanparams is not None:
        scan_e0 = scanparams.get('e0', None)
        if scan_e0 is not None:
            group.scan_e0 = float(scan_e0)
        scan_elem = scanparams.get('element', None)
        if scan_elem is not None:
            group.element = scan_elem
        scan_edge = scanparams.get('edge', None)
        if scan_edge is not None:
            group.edge = scan_edge
        scan_type = scanparams.get('scantype', None)
        if scan_type is not None:
            group.scan_type = scan_type

    ocrs, icrs = [], []
    ctime = None
    for attrname in dir(xdi):
        if attrname.lower() == 'counttime':
            ctime = getattr(xdi, attrname)
    if ctime is None:
        try:
            ctime = xdi.TSCALER / 5.e7
        except AttributeError:
            ctime = 1.0

    is_old_xsp3 = any(['13QX4' in a[1] for a in xdi.attrs['column'].items()])

    dtc_mode = 'icr/ocr'
    for i in range(MAX_MCAS):
        mca = "mca%i" % (i+1)
        ocr    = getattr(xdi, 'OutputCounts_%s' % mca, None)
        clock  = getattr(xdi, 'Clock_%s'        % mca, None)
        icr    = getattr(xdi, 'InputCounts_%s'  % mca, None)
        dtfact = getattr(xdi, 'DTFactor_%s'     % mca, None)
        resets = getattr(xdi, 'ResetTicks_%s'   % mca, None)
        allevt = getattr(xdi, 'AllEvent_%s'     % mca, None)
        if (ocr is None and icr is None and clock is None and
            dtfact is None and resets is None):
            nmca = i
            break
        if ocr is None:
            ocr = ctime
        ocr = ocr/ctime
        if icr is not None:
            icr = icr/ctime

        # Get ICR from one of several alternatives:
        # 1. InputCounts_mca was given
        # 2. DTFactor_mca was given
        # 3. ResetTicks_mca and AllEvets_mca were given
        # 4. Use "known values" for tau
        if icr is None:
            if dtfact is not None:
                icr = ocr * dtfact
                dtc_mode = 'dtfactor'
            elif (clock is not None and
                  resets is not None and
                  allevt is not None):
                dtfact = clock/(clock - (6*allevt + resets))
                icr = ocr * dtfact
                dtc_mode = 'resets'
        # finally estimate from measured values of tau:
        if icr is None:
            icr = 1.0*ocr
            if is_old_xsp3:
                icr = estimate_icr(ocr*1.00, XSPRESS3_TAU, niter=7)
        ocrs.append(ocr)
        icrs.append(icr)

    group.dtc_mode =  dtc_mode
    labels = []
    sums = {}
    for i, arrname in enumerate(xdi.array_labels):
        dat = getattr(xdi, arrname)
        if arrname.lower() == 'data':
            arrname = '_data'
        aname = sumname = rawname = arrname.lower()
        if ('_mca' in aname and
            'outputcounts' not in aname and
            'dtfactor' not in aname and
            'clock' not in aname):
            sumname, imca = sumname.split('_mca')
            if bad is not None and int(imca) in bad:
                sumname = sumname + '_bad'
            datraw = dat*1.0
            rawname = sumname + '_nodtc'
            imca = int(imca) - 1
            dat   = dat * icrs[imca]/ ocrs[imca]
            if any(np.isnan(dat)):
                nan_pts = np.where(np.isnan(dat))[0]
                dat[nan_pts] = datraw[nan_pts]

        setattr(group, aname, dat)
        if sumname not in labels:
            labels.append(sumname)
            sums[sumname] = dat
            if rawname != sumname:
                sums[rawname] = datraw
                if rawname not in labels:
                    labels.append(rawname)

        else:
            sums[sumname] = sums[sumname] + dat
            if rawname != sumname:
                sums[rawname] = sums[rawname] + datraw

    for name, dat in sums.items():
        if not hasattr(group, name):
            setattr(group, name, dat)

    for arrname in xdi.array_labels:
        sname = arrname.lower()
        if sname not in labels:
            labels.append(sname)

    data = []
    for name in labels:
        data.append(getattr(group, name))
    group.data = np.array(data)
    for imca in range(nmca):
        setattr(group, 'ocr_mca%i' % (imca+1), ocrs[imca])
        setattr(group, 'icr_mca%i' % (imca+1), icrs[imca])
    group.array_labels = labels
    return group


DTC_header = '''# XDI/1.1  Epics StepScan File/2.0
# Beamline.name:  13-ID-E, GSECARS
# Monochromator.name:  %(mono_cut)s, LN2 Cooled
# Monochromator.dspacing:  %(mono_dspace)s
# Facility.name: APS
# Facility.xray_source: 3.6 cm undulator
# Detectors.i0:  20cm ion chamber, He
# Detectors.ifluor:  Si SDD Vortex ME-4, 4 elements
# Detectors.ifluor_electronics:  Quantum Xspress3 3.1.10'''

def is_GSEXDI(filename):
    """test if file is GSE XDI data file
    reads only the first line of file
    """
    line1 = read_textfile(filename, size=40).split('\n')[0]
    return (line1.startswith('#XDI/1') and 'Epics StepScan File' in line1)

def gsexdi_deadtime_correct(fname, channelname, subdir='DT_Corrected',
                            bad=None):
    """convert GSE XDI fluorescence XAFS scans to dead time corrected files"""
    if not is_GSEXDI(fname):
        print("'%s' is not a GSE XDI scan file\n" % fname)
        return

    out = Group()
    out.orig_filename = fname
    try:
        xdi = read_gsexdi(fname, bad=bad)
    except:
        print('Could not read XDI file ', fname)
        return

    for attr in ('energy', 'i0', 'i1', 'i2', 'tscaler',
                 'counttime',  'scan_start_time', 'scan_end_time'):
        if hasattr(xdi, attr):
            setattr(out, attr, getattr(xdi, attr))

    # some scans may not record separate counttime, but TSCALER
    # is clock ticks for a 50MHz clock
    if not hasattr(out, 'counttime'):
        out.counttime = xdi.tscaler * 2.e-8

    if hasattr(xdi, 'energy_readback'):
        out.energy = xdi.energy_readback

    arrname = None
    channelname = channelname.lower().replace(' ', '_')

    for arr in xdi.array_labels:
        if arr.lower().startswith(channelname):
            arrname = arr
            break
    if arrname is None:
        print('Cannot find Channel %s in file %s '% (channelname, fname))
        return

    out.ifluor = getattr(xdi, arrname)
    out.ifluor_raw  = getattr(xdi, arrname)
    arrname_raw = arrname + '_nodtc'
    if arrname_raw  in xdi.array_labels:
        out.ifluor_raw  = getattr(xdi, arrname_raw)

    out.mufluor = out.ifluor / out.i0
    TINY  = 2.e-20
    if hasattr(out, 'i1') or hasattr(out, 'itrans'):
        i1 = getattr(out, 'i1', None)
        if i1 is None:
            i1 = getattr(out, 'itrans', None)
        if i1 is not None:
            i1[np.isnan(i1)] = TINY
            i1 = i1 / out.i0
            i1[np.where(i1<TINY)] = TINY
            out.mutrans = -np.log(i1)


    npts   = len(out.i0)
    col0_name = xdi.array_labels[0].lower()
    col0_units = None
    col0_data  = xdi.data[0, :]
    if col0_name == 'energy':
        col0_data  = out.energy
        col0_units = 'eV'

    buff =  ['# XDI/1.0  GSE/1.0']

    header = {}
    hgroups = ['beamline', 'facility', 'mono', 'undulator', 'detectors',
               'scaler', 'detectorstage', 'samplestage', 'scan', 'scanparameters']
    hskip = ['scanparameters.end', 'scanparameters.start']
    for agroup in hgroups:
        attrs = xdi._xdi.attrs.get(agroup, {})
        if agroup == 'mono': agroup = 'monochromator'
        header[agroup] = {}
        for sname in sorted(attrs.keys()):
            if "%s.%s" %( agroup, sname) not in hskip:
                header[agroup][sname] = attrs[sname]


    header['facility']['name'] = 'APS'
    header['facility']['xray_source'] = '3.6 cm undulator'
    header['beamline']['name'] = '13-ID-E, GSECARS'

    header['detectors']['i0'] = '20cm ion chamber, He'
    header['detectors']['ifluor'] = 'Si SDD Vortex ME-4, 4 elements'
    header['detectors']['ifluor_electronics'] = 'Quantum Xspress3 3.1.10'

    mono_cut = 'Si(111)'
    if xdi.mono_dspacing < 2:
        mono_cut = 'Si(311)'
    header['monochromator']['name'] = "%s, LN2 cooled"  % mono_cut

    out_arrays = {}
    out_arrays[col0_name]  = (col0_name, col0_units)
    out_arrays['mufluor'] = ('mufluor', None)
    if hasattr(out, 'i1'):
        out_arrays['mutrans'] = ('mutrans', None)

    out_arrays['ifluor']  = ('ifluor', '# deadtime-corrected')
    out_arrays['ifluor_raw'] = ('ifluor_raw', '# not deadtime-corrected')
    out_arrays['i0'] = ('i0', None)

    if hasattr(out, 'i1'):
        out_arrays['itrans'] = ('i1', None)
    if hasattr(out, 'i2'):
        out_arrays['irefer'] = ('i2', None)

    if hasattr(out, 'counttime'):
        out_arrays['counttime'] = ('counttime', 'sec')

    arrlabel = []
    for iarr, aname in enumerate(out_arrays):
        lab = "%12s " % aname
        if iarr == 0: lab = "%11s " % aname
        arrlabel.append(lab)
        extra = out_arrays[aname][1]
        if extra is None: extra = ''
        buff.append("# Column.%i: %s %s" % (iarr+1, aname, extra))

    arrlabel = '#%s' % (' '.join(arrlabel))
    ncol = len(out_arrays)

    for family, fval in header.items():
        for attr, val in fval.items():
            buff.append("# %s.%s: %s" % (family.title(), attr, val))


    buff.append("# ///")
    for comment in bytes2str(xdi._xdi.comments).split('\n'):
        c = comment.strip()
        if len(c) > 0:
            buff.append('# %s' % c)
    buff.extend(["# summed %s fluorescence data from %s" % (channelname, fname),
                 "# Dead-time correction applied",
                 "#"+ "-"*78,   arrlabel])

    efmt = "%11.4f"
    ffmt = "%13.7f"
    gfmt = "%13.7g"
    for i in range(npts):
        dline = ["", efmt % col0_data[i], ffmt % out.mufluor[i]]

        if hasattr(out, 'i1'):
            dline.append(ffmt % out.mutrans[i])

        dline.extend([gfmt % out.ifluor[i],
                      gfmt % out.ifluor_raw[i],
                      gfmt % out.i0[i]])

        if hasattr(out, 'i1'):
            dline.append(gfmt % out.i1[i])
        if hasattr(out, 'i2'):
            dline.append(gfmt % out.i2[i])
        if hasattr(out, 'counttime'):
            dline.append(gfmt % out.counttime[i])

        buff.append(" ".join(dline))
    ofile = fname[:]
    if ofile.startswith('..'):
        ofile = ofile[3:]
    ofile = ofile.replace('.', '_') + '.dat'
    ofile = os.path.join(subdir, ofile)
    if not os.path.exists(subdir):
        os.mkdir(subdir)
    try:
       fout = open(ofile, 'w', encoding=sys.getdefaultencoding())
       fout.write("\n".join(buff))
       fout.close()
       print("wrote %s, npts=%i, channel='%s'" % (ofile, npts, channelname))
    except:
       print("could not open / write to output file %s" % ofile)

    return out
